import functools
from typing import Any, Callable, Dict, List, TypeVar

CallableType = TypeVar("CallableType", bound=Callable)


_cache_list: List[Dict[Any, Any]] = []


def clear_cache() -> None:
    for c in _cache_list:
        c.clear()


def unsafe_cache(maxsize: int) -> Callable[[CallableType], CallableType]:
    """Cache decorator that relies object IDs when arguments are unhashable. Makes the
    very strong assumption of not only immutability, but that unhashable types don't go
    out of scope."""

    _cache_list.append({})
    local_cache = _cache_list[-1]

    def inner(f: CallableType) -> CallableType:
        @functools.wraps(f)
        def wrapped_f(*args, **kwargs):
            key = tuple(unsafe_hash(arg) for arg in args) + tuple(
                ("__kwarg__", k, unsafe_hash(v)) for k, v in kwargs.items()
            )

            if key in local_cache:
                return local_cache[key]

            out = f(*args, **kwargs)
            local_cache[key] = out
            if len(local_cache) > maxsize:
                local_cache.pop(next(iter(local_cache)))
            return out

        return wrapped_f  # type: ignore

    return inner


def unsafe_hash(obj: Any) -> Any:
    try:
        return hash(obj)
    except TypeError:
        return id(obj)



"""Tests initializing flax modules directly via tyro."""

from typing import cast

import jax
import pytest
from flax import linen as nn
from helptext_utils import get_helptext_with_checks
from jax import numpy as jnp

import tyro


class Classifier(nn.Module):
    layers: int
    units: int
    output_dim: int

    @nn.compact
    def __call__(self, x: jnp.ndarray) -> jnp.ndarray:  # type: ignore
        for i in range(self.layers - 1):
            x = nn.Dense(
                self.units,
                kernel_init=nn.initializers.kaiming_normal(),
            )(x)
            x = nn.relu(x)

        x = nn.Dense(
            self.output_dim,
            kernel_init=nn.initializers.xavier_normal(),
        )(x)
        x = nn.sigmoid(x)
        return x


def test_ok():
    network = tyro.cli(
        Classifier,
        args=[
            "--layers",
            "3",
            "--units",
            "8",
            "--output-dim",
            "3",
        ],
    )

    x = jnp.zeros((10, 4))
    params = network.init(jax.random.PRNGKey(0), x)
    assert cast(jax.Array, network.apply(params, x)).shape == (10, 3)

    helptext = get_helptext_with_checks(Classifier)
    assert "parent" not in helptext
    assert "name" not in helptext


def test_missing():
    with pytest.raises(SystemExit):
        tyro.cli(
            Classifier,
            args=[
                "--layers",
                "3",
                "--units",
                "8",
            ],
        )



from __future__ import annotations

import json
from typing import Any, Dict, Union

import numpy as np
from typing_extensions import Annotated, Literal, get_args

import tyro

json_constructor_spec = tyro.constructors.PrimitiveConstructorSpec(
    nargs=1,
    metavar="JSON",
    instance_from_str=lambda args: json.loads(args[0]),
    is_instance=lambda x: isinstance(x, dict),
    str_from_instance=lambda x: [json.dumps(x)],
)


def test_custom_primitive_registry():
    """Test that we can use a custom primitive registry to parse a custom type."""
    primitive_registry = tyro.constructors.ConstructorRegistry()

    @primitive_registry.primitive_rule
    def json_dict_spec(
        type_info: tyro.constructors.PrimitiveTypeInfo,
    ) -> tyro.constructors.PrimitiveConstructorSpec | None:
        if not (
            type_info.type_origin is dict and get_args(type_info.type) == (str, Any)
        ):
            return None
        return json_constructor_spec

    def main(x: Dict[str, Any]) -> Dict[str, Any]:
        return x

    with primitive_registry:
        assert tyro.cli(main, args=["--x", '{"a": 1}']) == {"a": 1}

    def main_with_default(x: Dict[str, Any] = {"hello": 5}) -> Dict[str, Any]:
        return x

    with primitive_registry:
        assert tyro.cli(main_with_default, args=[]) == {"hello": 5}
        assert tyro.cli(main_with_default, args=["--x", '{"a": 1}']) == {"a": 1}


def test_custom_primitive_annotated():
    """Test that we can use typing.Annotated to specify custom constructors."""

    def main(x: Annotated[Dict[str, Any], json_constructor_spec]) -> Dict[str, Any]:
        return x

    assert tyro.cli(main, args=["--x", '{"a": 1}']) == {"a": 1}


def test_custom_primitive_union():
    """Test that we can use typing.Annotated to specify custom constructors."""

    def main(
        x: Union[int, Annotated[Dict[str, Any], json_constructor_spec]],
    ) -> Union[int, Dict[str, Any]]:
        return x

    assert tyro.cli(main, args=["--x", "3"]) == 3
    assert tyro.cli(main, args=["--x", '{"a": 1}']) == {"a": 1}


def _construct_array(
    values: tuple[float, ...], dtype: Literal["float32", "float64"] = "float64"
) -> np.ndarray:
    return np.array(
        values,
        dtype={"float32": np.float32, "float64": np.float64}[dtype],
    )


def test_custom_constructor_numpy() -> None:
    def main(
        array: Annotated[np.ndarray, tyro.conf.arg(constructor=_construct_array)],
    ) -> np.ndarray:
        return array

    assert np.allclose(
        tyro.cli(main, args="--array.values 1 2 3 4 5".split(" ")),
        np.array([1, 2, 3, 4, 5], dtype=np.float64),
    )
    assert (
        tyro.cli(
            main, args="--array.values 1 2 3 4 5 --array.dtype float32".split(" ")
        ).dtype
        == np.float32
    )



"""Choosing Base Configs

One common pattern is to have a set of "base" configurations, which can be
selected from and then overridden.

This is often implemented with a set of configuration files (e.g., YAML files).
With :mod:`tyro`, we can instead define each base configuration as a separate
Python object.

After creating the base configurations, we can use the CLI to select one of
them and then override (existing) or fill in (missing) values.

The helper function used here, :func:`tyro.extras.overridable_config_cli()`, is
a lightweight wrapper over :func:`tyro.cli()` and its Union-based subcommand
syntax.


Usage:

    # Overall helptext:
    python ./03_choosing_base_configs.py --help

    # The "small" subcommand:
    python ./03_choosing_base_configs.py small --help
    python ./03_choosing_base_configs.py small --seed 94720

    # The "big" subcommand:
    python ./03_choosing_base_configs.py big --help
    python ./03_choosing_base_configs.py big --seed 94720
"""

from dataclasses import dataclass
from typing import Callable, Literal

from torch import nn

import tyro


@dataclass(frozen=True)
class ExperimentConfig:
    # Dataset to run experiment on.
    dataset: Literal["mnist", "imagenet-50"]

    # Model size.
    num_layers: int
    units: int

    # Batch size.
    batch_size: int

    # Total number of training steps.
    train_steps: int

    # Random seed.
    seed: int

    # Not specifiable via the commandline.
    activation: Callable[[], nn.Module]


# We could also define this library using separate YAML files (similar to
# `config_path`/`config_name` in Hydra), but staying in Python enables seamless
# type checking + IDE support.
default_configs = {
    "small": (
        "Small experiment.",
        ExperimentConfig(
            dataset="mnist",
            batch_size=2048,
            num_layers=4,
            units=64,
            train_steps=30_000,
            seed=0,
            activation=nn.ReLU,
        ),
    ),
    "big": (
        "Big experiment.",
        ExperimentConfig(
            dataset="imagenet-50",
            batch_size=32,
            num_layers=8,
            units=256,
            train_steps=100_000,
            seed=0,
            activation=nn.GELU,
        ),
    ),
}
if __name__ == "__main__":
    config = tyro.extras.overridable_config_cli(default_configs)
    print(config)



"""JAX/Flax Integration

If you use `flax.linen <https://github.com/google/flax>`_, modules can be instantiated
directly from :func:`tyro.cli()`.

Usage:

    python ./02_flax.py --help
    python ./02_flax.py --model.layers 4
"""

from flax import linen as nn
from jax import numpy as jnp

import tyro


class Classifier(nn.Module):
    layers: int
    """Layers in our network."""
    units: int = 32
    """Hidden unit count."""
    output_dim: int = 10
    """Number of classes."""

    @nn.compact
    def __call__(self, x: jnp.ndarray) -> jnp.ndarray:  # type: ignore
        for i in range(self.layers - 1):
            x = nn.Dense(
                self.units,
                kernel_init=nn.initializers.kaiming_normal(),
            )(x)
            x = nn.relu(x)

        x = nn.Dense(
            self.output_dim,
            kernel_init=nn.initializers.xavier_normal(),
        )(x)
        x = nn.sigmoid(x)
        return x


def train(model: Classifier, num_iterations: int = 1000) -> None:
    """Train a model.

    Args:
        model: Model to train.
        num_iterations: Number of training iterations.
    """
    print(f"{model=}")
    print(f"{num_iterations=}")


if __name__ == "__main__":
    tyro.cli(train)



import builtins
import dataclasses
from typing import Any

import pytest


# https://stackoverflow.com/questions/60227582/making-a-python-test-think-an-installed-package-is-not-available
@pytest.fixture
def hide_optional_packages(monkeypatch: Any) -> None:
    import_orig = builtins.__import__

    def mocked_import(name, *args, **kwargs):
        if name in ("attr", "pydantic", "flax", "omegaconf"):
            raise ImportError()
        return import_orig(name, *args, **kwargs)

    monkeypatch.setattr(builtins, "__import__", mocked_import)


def test_missing_optional_packages(hide_optional_packages: Any) -> None:
    @dataclasses.dataclass
    class Extra:
        a: int = 3
        b: str = "5"

    def main(x: int, extra: Extra) -> int:
        del extra
        return x

    import tyro

    assert tyro.cli(main, args=["--x", "5"]) == 5



"""Type-safe, human-readable serialization helpers for dataclasses."""

from __future__ import annotations

import dataclasses
import enum
import functools
from typing import IO, TYPE_CHECKING, Any, Optional, Set, Type, TypeVar, Union

if TYPE_CHECKING:
    # Since serialization functionality is deprecated, the yaml dependency is optional.
    import yaml

from typing_extensions import get_args, get_origin

from .. import _resolver
from ..constructors._struct_spec import MISSING

ENUM_YAML_TAG_PREFIX = "!enum:"
DATACLASS_YAML_TAG_PREFIX = "!dataclass:"
MISSING_YAML_TAG_PREFIX = "!missing"

DataclassType = TypeVar("DataclassType")


def _get_contained_special_types_from_type(
    cls: Type[Any],
    _parent_contained_dataclasses: Optional[Set[Type[Any]]] = None,
) -> Set[Type[Any]]:
    """Takes a dataclass type, and recursively searches its fields for dataclass or enum
    types."""
    assert _resolver.is_dataclass(cls)
    parent_contained_dataclasses = (
        set()
        if _parent_contained_dataclasses is None
        else _parent_contained_dataclasses
    )

    cls = _resolver.unwrap_annotated(cls)
    cls, type_from_typevar = _resolver.resolve_generic_types(cls)

    contained_special_types = {cls}

    def handle_type(typ: Type[Any]) -> Set[Type[Any]]:
        # Handle dataclasses.
        if _resolver.is_dataclass(typ) and typ not in parent_contained_dataclasses:
            return _get_contained_special_types_from_type(
                typ,
                _parent_contained_dataclasses=contained_special_types
                | parent_contained_dataclasses,
            )

        # Handle enums.
        elif isinstance(typ, enum.EnumMeta):
            return {typ}

        # Handle Union, Annotated, List, etc. No-op when there are no args.
        return functools.reduce(set.union, map(handle_type, get_args(typ)), set())

    # Handle generics.
    for typ in type_from_typevar.values():
        contained_special_types |= handle_type(typ)

    if cls in parent_contained_dataclasses:
        return contained_special_types

    # Handle fields.
    for field in _resolver.resolved_fields(cls):  # type: ignore
        assert not isinstance(field.type, str)
        contained_special_types |= handle_type(field.type)

    # Handle subclasses.
    for subclass in cls.__subclasses__():
        contained_special_types |= handle_type(subclass)

    return contained_special_types


def _make_loader(cls: Type[Any]) -> Type[yaml.Loader]:
    import yaml

    class DataclassLoader(yaml.Loader):
        pass

    # Design Q: do we want to support multiple dataclass types with the same name?
    # - Why yes: rudimentary support for this is easy.
    # - Why no: might cause confusing error messages? Using all possible context cues
    # for this is also hard.
    #
    # => let's just keep things simple, assert uniqueness for now. Easier to add new
    # features later than remove them.

    contained_types = list(_get_contained_special_types_from_type(cls))
    contained_type_names = list(map(lambda cls: cls.__name__, contained_types))
    assert len(set(contained_type_names)) == len(contained_type_names), (
        "Contained dataclass type names must all be unique, but got"
        f" {contained_type_names}"
    )

    def make_dataclass_constructor(typ: Type[Any]):
        return lambda loader, node: typ(**loader.construct_mapping(node))

    def make_enum_constructor(typ: Type[enum.Enum]):
        return lambda loader, node: typ[loader.construct_python_str(node)]

    for typ, name in zip(contained_types, contained_type_names):
        if dataclasses.is_dataclass(typ):
            DataclassLoader.add_constructor(
                tag=DATACLASS_YAML_TAG_PREFIX + name,
                constructor=make_dataclass_constructor(typ),
            )
        elif issubclass(typ, enum.Enum):
            DataclassLoader.add_constructor(
                tag=ENUM_YAML_TAG_PREFIX + name,
                constructor=make_enum_constructor(typ),
            )
        else:
            assert False

    DataclassLoader.add_constructor(
        tag=MISSING_YAML_TAG_PREFIX,
        constructor=lambda *_unused: MISSING,  # type: ignore
    )

    return DataclassLoader


def _make_dumper(instance: Any) -> Type[yaml.Dumper]:
    import yaml

    class DataclassDumper(yaml.Dumper):
        def ignore_aliases(self, data):
            return super().ignore_aliases(data) or data is MISSING

    contained_types = list(_get_contained_special_types_from_type(type(instance)))
    contained_type_names = list(map(lambda cls: cls.__name__, contained_types))

    # Note: this is currently a stricter than necessary assert.
    assert len(set(contained_type_names)) == len(contained_type_names), (
        "Contained dataclass/enum names must all be unique, but got"
        f" {contained_type_names}"
    )

    def make_representer(name: str):
        def representer(dumper: DataclassDumper, data: Any) -> yaml.Node:
            if dataclasses.is_dataclass(data):
                return dumper.represent_mapping(
                    tag=DATACLASS_YAML_TAG_PREFIX + name,
                    mapping={
                        field.name: getattr(data, field.name)
                        for field in dataclasses.fields(data)
                        if field.init
                    },
                )
            elif isinstance(data, enum.Enum):
                return dumper.represent_scalar(
                    tag=ENUM_YAML_TAG_PREFIX + name, value=data.name
                )
            assert False

        return representer

    for typ, name in zip(contained_types, contained_type_names):
        DataclassDumper.add_representer(typ, make_representer(name))

    DataclassDumper.add_representer(
        type(MISSING),
        lambda dumper, data: dumper.represent_scalar(
            tag=MISSING_YAML_TAG_PREFIX, value=""
        ),
    )
    return DataclassDumper


def from_yaml(
    cls: Type[DataclassType],
    stream: Union[str, IO[str], bytes, IO[bytes]],
) -> DataclassType:
    """Re-construct a dataclass instance from a yaml-compatible string, which should be
    generated from :func:`tyro.extras.to_yaml()`.

    .. warning::

        **Deprecated.** Serialization functionality is stable but deprecated.
        It may be removed in a future version of :code:`tyro`.

    As a secondary feature aimed at enabling the use of :func:`tyro.cli` for general
    configuration use cases, we also introduce functions for human-readable dataclass
    serialization: :func:`tyro.extras.from_yaml` and :func:`tyro.extras.to_yaml` attempt
    to strike a balance between flexibility and robustness — in contrast to naively
    dumping or loading dataclass instances (via pickle, PyYAML, etc), explicit type
    references enable custom tags that are robust against code reorganization and
    refactor, while a PyYAML backend enables serialization of arbitrary Python objects.

    Args:
        cls: Type to reconstruct.
        stream: YAML to read from.

    Returns:
        Instantiated dataclass.
    """
    import yaml

    out = yaml.load(stream, Loader=_make_loader(cls))
    origin_cls = get_origin(cls)
    assert isinstance(out, origin_cls if origin_cls is not None else cls)
    return out


def to_yaml(instance: Any) -> str:
    """Serialize a dataclass; returns a yaml-compatible string that can be deserialized
    via :func:`tyro.extras.from_yaml()`.

    .. warning::

        **Deprecated.** Serialization functionality is stable but deprecated.
        It may be removed in a future version of :code:`tyro`.

    As a secondary feature aimed at enabling the use of :func:`tyro.cli` for general
    configuration use cases, we also introduce functions for human-readable dataclass
    serialization: :func:`tyro.extras.from_yaml` and :func:`tyro.extras.to_yaml` attempt
    to strike a balance between flexibility and robustness — in contrast to naively
    dumping or loading dataclass instances (via pickle, PyYAML, etc), explicit type
    references enable custom tags that are robust against code reorganization and
    refactor, while a PyYAML backend enables serialization of arbitrary Python objects.

    Args:
        instance: Dataclass instance to serialize.

    Returns:
        YAML string.
    """
    import yaml

    return "# tyro YAML.\n" + yaml.dump(instance, Dumper=_make_dumper(instance))



"""Generate a Python 3.11 version of tests. This will use imports from `typing` instead
of `typing_extensions`, and replace Union[A, B] types with A | B."""

import pathlib
import subprocess
from concurrent.futures import ThreadPoolExecutor


def generate_from_path(test_path: pathlib.Path) -> None:
    content = test_path.read_text()
    content = content.replace("typing_extensions", "typing")

    if "ReadOnly" in content:
        content = content.replace("ReadOnly,", "", 1)
        content = content.replace(
            "\nfrom typing import ",
            "\nfrom typing_extensions import ReadOnly\nfrom typing import ",
        )

    while "Union[" in content:
        new_content, _, b = content.partition("Union[")

        if b.strip()[0] == '"':
            break  # Don't bother with forward references!

        new_content_parts = [new_content]

        bracket_count = 0
        for i, char in enumerate(b):
            if char == "[":
                bracket_count += 1
            elif char == "]":
                bracket_count -= 1

            if char == "," and bracket_count == 0:
                new_content_parts.append("|")
            elif bracket_count == -1:
                while new_content_parts[-1] in (" ", "|"):
                    new_content_parts.pop(-1)
                new_content_parts.append(b[i + 1 :])
                break
            elif char != "\n":
                new_content_parts.append(char)

        content = "".join(new_content_parts)

    out_path = pathlib.Path(__file__).absolute().parent / (
        test_path.stem + "_generated.py"
    )
    out_path.write_text(content)

    subprocess.run(["ruff", "format", str(out_path)], check=True)
    subprocess.run(["ruff", "check", "--fix", str(out_path)], check=True)


with ThreadPoolExecutor(max_workers=8) as executor:
    list(
        executor.map(
            generate_from_path,
            pathlib.Path(__file__).absolute().parent.parent.glob("test_*.py"),
        )
    )



"""Attrs Integration

In addition to standard dataclasses, :func:`tyro.cli()` also supports
`attrs <https://www.attrs.org/>`_ classes.

Usage:

    python ./07_attrs.py --help
    python ./07_attrs.py --field1 hello
    python ./07_attrs.py --field1 hello --field2 5
"""

import attr

import tyro


@attr.s
class Args:
    """Description.
    This should show up in the helptext!"""

    field1: str = attr.ib()
    """A string field."""

    field2: int = attr.ib(factory=lambda: 5)
    """A required integer field."""


if __name__ == "__main__":
    args = tyro.cli(Args)
    print(args)



"""Configuration via typing.Annotated[]

The :mod:`tyro.conf` module contains utilities that can be used in conjunction
with :py:data:`typing.Annotated` to configure command-line interfaces beyond
what is expressible via static type annotations.

Features here are supported, but generally unnecessary and should be used sparingly.

Usage:

    python ./09_conf.py --help
    python ./09_conf.py 5 --boolean True
"""

import dataclasses

from typing_extensions import Annotated

import tyro


@dataclasses.dataclass
class Args:
    # A numeric field parsed as a positional argument.
    positional: tyro.conf.Positional[int]

    # A boolean field with flag conversion turned off.
    boolean: tyro.conf.FlagConversionOff[bool] = False

    # A numeric field that can't be changed via the CLI.
    fixed: tyro.conf.Fixed[int] = 5

    # A field with manually overridden properties.
    manual: Annotated[
        str,
        tyro.conf.arg(
            name="renamed",
            metavar="STRING",
            help="A field with manually overridden properties!",
        ),
    ] = "Hello"


if __name__ == "__main__":
    print(tyro.cli(Args))



"""Sequenced Subcommands

Multiple unions over struct types are populated using a series of subcommands.

Usage:

    # We apply the :class:`tyro.conf.ConsolidateSubcommandArgs` flag. This
    # pushes all arguments to the end of the command:
    python ./03_multiple_subcommands.py --help
    python ./03_multiple_subcommands.py dataset:mnist --help
    python ./03_multiple_subcommands.py dataset:mnist optimizer:adam --help
    python ./03_multiple_subcommands.py dataset:mnist optimizer:adam --optimizer.learning-rate 3e-4 --dataset.binary
"""

from __future__ import annotations

import dataclasses
from typing import Literal

import tyro

# Possible dataset configurations.


@dataclasses.dataclass
class Mnist:
    binary: bool = False
    """Set to load binary version of MNIST dataset."""


@dataclasses.dataclass
class ImageNet:
    subset: Literal[50, 100, 1000]
    """Choose between ImageNet-50, ImageNet-100, ImageNet-1000, etc."""


# Possible optimizer configurations.


@dataclasses.dataclass
class Adam:
    learning_rate: float = 1e-3
    betas: tuple[float, float] = (0.9, 0.999)


@dataclasses.dataclass
class Sgd:
    learning_rate: float = 3e-4


# Train script.


def train(
    dataset: Mnist | ImageNet = Mnist(),
    optimizer: Adam | Sgd = Adam(),
) -> None:
    """Example training script.

    Args:
        dataset: Dataset to train on.
        optimizer: Optimizer to train with.

    Returns:
        None:
    """
    print(dataset)
    print(optimizer)


if __name__ == "__main__":
    tyro.cli(train, config=(tyro.conf.ConsolidateSubcommandArgs,))



import sys
from typing import List

collect_ignore_glob: List[str] = []

if not sys.version_info >= (3, 8):
    collect_ignore_glob.append("*min_py38*.py")

if not sys.version_info >= (3, 9):
    collect_ignore_glob.append("*min_py39*.py")

if not sys.version_info >= (3, 10):
    collect_ignore_glob.append("*min_py310*.py")
    collect_ignore_glob.append("*_min_py310_generated.py")

if not sys.version_info >= (3, 12):
    collect_ignore_glob.append("*min_py312*.py")

if not sys.version_info >= (3, 11):
    collect_ignore_glob.append("test_py311_generated/*.py")

if sys.version_info >= (3, 13):
    collect_ignore_glob.append("*_exclude_py313*.py")



"""Adapted from: https://github.com/brentyi/tyro/issues/183"""

from typing import NamedTuple, Set

from helptext_utils import get_helptext_with_checks
from pydantic import BaseModel, Field
from typing_extensions import Annotated

import tyro


class MyRange(NamedTuple):
    low: int
    high: int

    def __str__(self):
        return f"<{self.low}, {self.high}>"

    @staticmethod
    def tyro_constructor(
        range_str: Annotated[
            str,
            tyro.conf.arg(name=""),
        ],
    ):
        import re

        m = re.match("([0-9]+)(-([0-9]+))*", range_str)
        low = m[1]  # type: ignore
        high = low if not m[3] else m[3]  # type: ignore

        return MyRange(int(low), int(high))

    @staticmethod
    def tyro_constructor_set(
        range_str_set: Annotated[
            Set[str],
            tyro.conf.arg(name=""),
        ],
    ):
        return {MyRange.tyro_constructor(r) for r in range_str_set}


class MySpec(BaseModel):
    some_set: Set[int] = Field(
        default={1, 2, 3},
        description="Some set of integers",
        title="Some set",
    )

    some_string: str = Field(
        description="Some string without a default value.", title="SomeSTR"
    )

    here_comes_the_trouble: Annotated[
        Set[MyRange],
        tyro.conf.arg(constructor=MyRange.tyro_constructor_set),
    ] = Field(
        default={MyRange(0, 1024)},
        description="I would like this one in the same group as others",
        title="Please help",
    )


def add_spec(spec: MySpec) -> MySpec:
    return spec


def test_functionality() -> None:
    assert tyro.cli(
        add_spec, args=["--spec.some-set", "1", "2", "3", "--spec.some-string", "hello"]
    ) == MySpec(
        some_set={1, 2, 3},
        some_string="hello",
        here_comes_the_trouble={MyRange(0, 1024)},
    )
    assert tyro.cli(
        add_spec,
        args=[
            "--spec.some-set",
            "1",
            "2",
            "3",
            "--spec.some-string",
            "hello",
            "--spec.here-comes-the-trouble",
            "0-512",
        ],
    ) == MySpec(
        some_set={1, 2, 3},
        some_string="hello",
        here_comes_the_trouble={MyRange(0, 512)},
    )


def test_helptext() -> None:
    helptext = get_helptext_with_checks(add_spec)
    assert "spec options" in helptext
    assert "spec.here-comes-the-trouble-options" not in helptext



"""Utilities for resolving types and forward references."""

from __future__ import annotations

import collections.abc
import copy
import dataclasses
import inspect
import sys
import types
import warnings
from typing import (
    Any,
    Callable,
    ClassVar,
    Dict,
    List,
    Sequence,
    Set,
    Tuple,
    Type,
    TypeVar,
    Union,
    cast,
    overload,
)

from typing_extensions import (
    Annotated,
    Final,
    ForwardRef,
    Literal,
    Self,
    TypeAliasType,
    get_args,
    get_origin,
    get_type_hints,
)

from . import _unsafe_cache, conf
from ._singleton import MISSING_AND_MISSING_NONPROP
from ._typing import TypeForm

UnionType = getattr(types, "UnionType", Union)
"""Same as types.UnionType, but points to typing.Union for older versions of
Python. types.UnionType was added in Python 3.10, and is created when the `X |
Y` syntax is used for unions."""

TypeOrCallable = TypeVar("TypeOrCallable", TypeForm[Any], Callable)


@dataclasses.dataclass(frozen=True)
class TyroTypeAliasBreadCrumb:
    """A breadcrumb we can leave behind to track names of type aliases and
    `NewType` types. We can use type alias names to auto-populate
    subcommands."""

    name: str


def unwrap_origin_strip_extras(typ: TypeOrCallable) -> TypeOrCallable:
    """Returns the origin, ignoring typing.Annotated, of typ if it exists. Otherwise,
    returns typ."""
    typ = unwrap_annotated(typ)
    origin = get_origin(typ)

    if origin is not None:
        typ = origin

    return typ


def is_dataclass(cls: Union[TypeForm, Callable]) -> bool:
    """Same as `dataclasses.is_dataclass`, but also handles generic aliases."""
    return dataclasses.is_dataclass(unwrap_origin_strip_extras(cls))  # type: ignore


@_unsafe_cache.unsafe_cache(maxsize=1024)
def resolved_fields(cls: TypeForm) -> List[dataclasses.Field]:
    """Similar to dataclasses.fields(), but includes dataclasses.InitVar types and
    resolves forward references."""

    assert dataclasses.is_dataclass(cls)
    fields = []
    annotations = get_type_hints_with_backported_syntax(
        cast(Callable, cls), include_extras=True
    )
    for field in getattr(cls, "__dataclass_fields__").values():
        # Avoid mutating original field.
        field = copy.copy(field)

        # Resolve forward references.
        field.type = annotations[field.name]

        # Skip ClassVars.
        if get_origin(field.type) is ClassVar:
            continue

        # Unwrap InitVar types.
        if isinstance(field.type, dataclasses.InitVar):
            field.type = field.type.type

        fields.append(field)

    return fields


def is_namedtuple(cls: TypeForm) -> bool:
    return (
        hasattr(cls, "_fields")
        # `_field_types` was removed in Python >=3.9.
        # and hasattr(cls, "_field_types")
        and hasattr(cls, "_field_defaults")
    )


def type_from_typevar_constraints(typ: TypeOrCallable) -> TypeOrCallable:
    """Try to concretize a type from a TypeVar's bounds or constraints. Identity if
    unsuccessful."""
    if isinstance(typ, TypeVar):
        if typ.__bound__ is not None:
            # Try to infer type from TypeVar bound.
            return typ.__bound__
        elif len(typ.__constraints__) > 0:
            # Try to infer type from TypeVar constraints.
            return Union.__getitem__(typ.__constraints__)  # type: ignore
        else:
            return Any
    return typ


TypeOrCallableOrNone = TypeVar("TypeOrCallableOrNone", Callable, TypeForm[Any], None)


def resolve_newtype_and_aliases(
    typ: TypeOrCallableOrNone,
) -> TypeOrCallableOrNone:
    # Handle type aliases, eg via the `type` statement in Python 3.12.
    if isinstance(typ, TypeAliasType):
        return Annotated[
            (
                cast(Any, resolve_newtype_and_aliases(typ.__value__)),
                TyroTypeAliasBreadCrumb(typ.__name__),
            )
        ]

    # We'll unwrap NewType annotations here; this is needed before issubclass
    # checks!
    #
    # `isinstance(x, NewType)` doesn't work because NewType isn't a class until
    # Python 3.10, so we instead do a duck typing-style check.
    return_name = None
    while hasattr(typ, "__name__") and hasattr(typ, "__supertype__"):
        if return_name is None:
            return_name = getattr(typ, "__name__")
        typ = resolve_newtype_and_aliases(getattr(typ, "__supertype__"))

    if return_name is not None:
        typ = Annotated[(typ, TyroTypeAliasBreadCrumb(return_name))]  # type: ignore

    return cast(TypeOrCallableOrNone, typ)


@_unsafe_cache.unsafe_cache(maxsize=1024)
def narrow_subtypes(
    typ: TypeOrCallable,
    default_instance: Any,
) -> TypeOrCallable:
    """Type narrowing: if we annotate as Animal but specify a default instance of Cat,
    we should parse as Cat.

    This should generally only be applied to fields used as nested structures, not
    individual arguments/fields. (if a field is annotated as Union[int, str], and a
    string default is passed in, we don't want to narrow the type to always be
    strings!)"""

    typ = resolve_newtype_and_aliases(typ)

    try:
        potential_subclass = type(default_instance)

        if potential_subclass is type:
            # Don't narrow to `type`. This happens when the default instance is a class;
            # it doesn't really make sense to parse this case.
            return typ

        superclass = unwrap_annotated(typ)

        # For Python 3.10.
        if get_origin(superclass) is Union:
            return typ

        if superclass is Any or issubclass(potential_subclass, superclass):  # type: ignore
            if get_origin(typ) is Annotated:
                return Annotated[(potential_subclass,) + get_args(typ)[1:]]  # type: ignore
            typ = cast(TypeOrCallable, potential_subclass)
    except TypeError:
        # TODO: document where this TypeError can be raised, and reduce the amount of
        # code in it.
        pass

    return typ


def swap_type_using_confstruct(typ: TypeOrCallable) -> TypeOrCallable:
    """Swap types using the `constructor_factory` attribute from
    `tyro.conf.arg` and `tyro.conf.subcommand`. Runtime annotations are
    kept, but the type is swapped."""
    # Need to swap types.
    _, annotations = unwrap_annotated(typ, search_type="all")
    for anno in reversed(annotations):
        if (
            isinstance(
                anno,
                (
                    conf._confstruct._ArgConfig,
                    conf._confstruct._SubcommandConfig,
                ),
            )
            and anno.constructor_factory is not None
        ):
            return Annotated[(anno.constructor_factory(),) + annotations]  # type: ignore
    return typ


def narrow_collection_types(
    typ: TypeOrCallable, default_instance: Any
) -> TypeOrCallable:
    """TypeForm narrowing for containers. Infers types of container contents."""
    args = get_args(typ)
    origin = get_origin(typ)
    if args == (Any,) or (origin is tuple and args == (Any, Ellipsis)):
        typ = origin  # type: ignore

    if typ in (list, Sequence, collections.abc.Sequence) and isinstance(
        default_instance, list
    ):
        if len(default_instance) == 0:
            return typ
        typ = List.__getitem__(Union.__getitem__(tuple(map(type, default_instance))))  # type: ignore
    elif typ in (set, Sequence, collections.abc.Sequence) and isinstance(
        default_instance, set
    ):
        if len(default_instance) == 0:
            return typ
        typ = Set.__getitem__(Union.__getitem__(tuple(map(type, default_instance))))  # type: ignore
    elif typ in (tuple, Sequence, collections.abc.Sequence) and isinstance(
        default_instance, tuple
    ):
        if len(default_instance) == 0:
            return typ
        typ = Tuple.__getitem__(tuple(map(type, default_instance)))  # type: ignore
    return cast(TypeOrCallable, typ)


# `Final` and `ReadOnly` types are ignored in tyro.
try:
    # Can only import ReadOnly in typing_extensions>=4.9.0, which isn't
    # supported by Python 3.7.
    from typing_extensions import ReadOnly

    STRIP_WRAPPER_TYPES = {Final, ReadOnly}
except ImportError:
    STRIP_WRAPPER_TYPES = {Final}

MetadataType = TypeVar("MetadataType")


@overload
def unwrap_annotated(
    typ: TypeOrCallable,
    search_type: TypeForm[MetadataType],
) -> Tuple[TypeOrCallable, Tuple[MetadataType, ...]]: ...


@overload
def unwrap_annotated(
    typ: TypeOrCallable,
    search_type: Literal["all"],
) -> Tuple[TypeOrCallable, Tuple[Any, ...]]: ...


@overload
def unwrap_annotated(
    typ: TypeOrCallable,
    search_type: None = None,
) -> TypeOrCallable: ...


def unwrap_annotated(
    typ: TypeOrCallable,
    search_type: Union[TypeForm[MetadataType], Literal["all"], object, None] = None,
) -> Union[Tuple[TypeOrCallable, Tuple[MetadataType, ...]], TypeOrCallable]:
    """Helper for parsing typing.Annotated types.

    Examples:
    - int, int => (int, ())
    - Annotated[int, 1], int => (int, (1,))
    - Annotated[int, "1"], int => (int, ())
    """

    # Unwrap aliases defined using Python 3.12's `type` syntax.
    typ = resolve_newtype_and_aliases(typ)

    # `Final` and `ReadOnly` types are ignored in tyro.
    while get_origin(typ) in STRIP_WRAPPER_TYPES:
        typ = get_args(typ)[0]

    # Don't search for any annotations.
    if search_type is None:
        if not hasattr(typ, "__metadata__"):
            return typ
        else:
            return get_args(typ)[0]

    # Check for __tyro_markers__ from @configure.
    if hasattr(typ, "__tyro_markers__"):
        if search_type == "all":
            targets = getattr(typ, "__tyro_markers__")
        else:
            targets = tuple(
                x
                for x in getattr(typ, "__tyro_markers__")
                if isinstance(x, search_type)  # type: ignore
            )
    else:
        targets = ()

    assert isinstance(targets, tuple)
    if not hasattr(typ, "__metadata__"):
        return typ, targets  # type: ignore

    args = get_args(typ)
    assert len(args) >= 2

    # Look through metadata for desired metadata type.
    targets += tuple(
        x
        for x in targets + args[1:]
        if search_type == "all" or isinstance(x, search_type)  # type: ignore
    )

    # Check for __tyro_markers__ in unwrapped type.
    if hasattr(args[0], "__tyro_markers__"):
        targets += tuple(
            x
            for x in getattr(args[0], "__tyro_markers__")
            if search_type == "all" or isinstance(x, search_type)  # type: ignore
        )
    return args[0], targets  # type: ignore


class TypeParamResolver:
    param_assignments: List[Dict[TypeVar, TypeForm[Any]]] = []

    @classmethod
    def get_assignment_context(cls, typ: TypeOrCallable) -> TypeParamAssignmentContext:
        """Context manager for resolving type parameters."""
        typ, type_from_typevar = resolve_generic_types(typ)
        return TypeParamAssignmentContext(typ, type_from_typevar)

    @staticmethod
    def concretize_type_params(
        typ: TypeOrCallable, seen: set[Any] | None = None
    ) -> TypeOrCallable:
        """Apply type parameter assignments based on the current context."""

        if seen is None:
            seen = set()
        elif seen is not None and typ in seen:
            # Found a cycle. We don't (currently) support recursive types.
            return typ
        else:
            seen.add(typ)

        typ = resolve_newtype_and_aliases(typ)
        type_from_typevar = {}
        GenericAlias = getattr(types, "GenericAlias", None)
        while (
            GenericAlias is not None
            and isinstance(typ, GenericAlias)
            and len(getattr(typ, "__type_params__", ())) > 0
        ):
            for k, v in zip(typ.__type_params__, get_args(typ)):  # type: ignore
                type_from_typevar[k] = TypeParamResolver.concretize_type_params(
                    v, seen=seen
                )
            typ = typ.__value__  # type: ignore

        if len(type_from_typevar) == 0:
            return TypeParamResolver._concretize_type_params(typ, seen=seen)
        else:
            with TypeParamAssignmentContext(typ, type_from_typevar):
                return TypeParamResolver._concretize_type_params(typ, seen=seen)

    @staticmethod
    def _concretize_type_params(typ: TypeOrCallable, seen: set[Any]) -> TypeOrCallable:
        for type_from_typevar in reversed(TypeParamResolver.param_assignments):
            if typ in type_from_typevar:
                return type_from_typevar[typ]  # type: ignore

        origin = get_origin(typ)
        args = get_args(typ)
        if len(args) > 0:
            if origin is Annotated:
                args = args[:1]
            if origin is collections.abc.Callable:
                assert isinstance(args[0], list)
                args = tuple(args[0]) + args[1:]

            new_args_list = []
            for x in args:
                for type_from_typevar in reversed(TypeParamResolver.param_assignments):
                    if x in type_from_typevar:
                        x = type_from_typevar[x]
                        break
                new_args_list.append(x)

            new_args = tuple(
                TypeParamResolver.concretize_type_params(x, seen=seen)
                for x in new_args_list
            )

            # Standard generic aliases have a `copy_with()`!
            if origin is UnionType:
                return Union.__getitem__(new_args)  # type: ignore
            elif hasattr(typ, "copy_with"):
                # typing.List, typing.Dict, etc.
                return typ.copy_with(new_args)  # type: ignore
            else:
                # list[], dict[], etc.
                assert origin is not None
                return origin[new_args]

        return typ  # type: ignore


class TypeParamAssignmentContext:
    def __init__(
        self,
        origin_type: TypeOrCallable,
        type_from_typevar: Dict[TypeVar, TypeForm[Any]],
    ):
        # `Any` is needed for mypy...
        self.origin_type: Any = origin_type
        self.type_from_typevar = type_from_typevar

    def __enter__(self):
        TypeParamResolver.param_assignments.append(self.type_from_typevar)

    def __exit__(self, exc_type, exc_value, traceback):
        TypeParamResolver.param_assignments.pop()


@_unsafe_cache.unsafe_cache(maxsize=1024)
def narrow_union_type(typ: TypeOrCallable, default_instance: Any) -> TypeOrCallable:
    """Narrow union types.

    This is a shim for failing more gracefully when we we're given a Union type that
    doesn't match the default value.

    In this case, we raise a warning, then add the type of the default value to the
    union. Loosely motivated by: https://github.com/brentyi/tyro/issues/20
    """
    if get_origin(typ) is not Union:
        return typ

    options = get_args(typ)
    options_unwrapped = [unwrap_origin_strip_extras(o) for o in options]

    try:
        if default_instance not in MISSING_AND_MISSING_NONPROP and not any(
            isinstance_with_fuzzy_numeric_tower(default_instance, o) is not False
            for o in options_unwrapped
        ):
            warnings.warn(
                f"{type(default_instance)} does not match any type in Union:"
                f" {options_unwrapped}"
            )
            return Union.__getitem__(options + (type(default_instance),))  # type: ignore
    except TypeError:
        pass

    return typ


def isinstance_with_fuzzy_numeric_tower(
    obj: Any, classinfo: Type
) -> Union[bool, Literal["~"]]:
    """
    Enhanced version of isinstance() that returns:
    - True: if object is exactly of the specified type
    - "~": if object follows numeric tower rules but isn't exact type
    - False: if object is not of the specified type or numeric tower rules don't apply

    Examples:
    >>> enhanced_isinstance(3, int)       # Returns True
    >>> enhanced_isinstance(3, float)     # Returns "~"
    >>> enhanced_isinstance(True, int)    # Returns "~"
    >>> enhanced_isinstance(3, bool)      # Returns False
    >>> enhanced_isinstance(True, bool)   # Returns True
    """
    # Handle exact match first
    if isinstance(obj, classinfo):
        return True

    # Handle numeric tower cases
    if isinstance(obj, bool):
        if classinfo in (int, float, complex):
            return "~"
    elif isinstance(obj, int) and not isinstance(obj, bool):  # explicit bool check
        if classinfo in (float, complex):
            return "~"
    elif isinstance(obj, float):
        if classinfo is complex:
            return "~"

    return False


NoneType = type(None)


def resolve_generic_types(
    typ: TypeOrCallable,
) -> Tuple[TypeOrCallable, Dict[TypeVar, TypeForm[Any]]]:
    """If the input is a class: no-op. If it's a generic alias: returns the origin
    class, and a mapping from typevars to concrete types."""

    annotations: Tuple[Any, ...] = ()
    if get_origin(typ) is Annotated:
        # ^We need this `if` statement for an obscure edge case: when `cls` is a
        # function with `__tyro_markers__` set, we don't want/need to return
        # Annotated[func, markers].
        typ, annotations = unwrap_annotated(typ, "all")

    # Apply shims to convert from types.UnionType to typing.Union, list to
    # typing.List, etc.
    typ = resolve_newtype_and_aliases(typ)

    # We'll ignore NewType when getting the origin + args for generics.
    origin_cls = get_origin(typ)
    type_from_typevar: Dict[TypeVar, TypeForm[Any]] = {}

    # Support typing.Self.
    # We'll do this by pretending that `Self` is a TypeVar...
    if hasattr(typ, "__self__"):
        self_type = getattr(typ, "__self__")
        if inspect.isclass(self_type):
            type_from_typevar[cast(TypeVar, Self)] = self_type  # type: ignore
        else:
            type_from_typevar[cast(TypeVar, Self)] = self_type.__class__  # type: ignore

    if (
        # Apply some heuristics for generic types. Should revisit this.
        origin_cls is not None
        and hasattr(origin_cls, "__parameters__")
        and hasattr(origin_cls.__parameters__, "__len__")
    ):
        typevars = origin_cls.__parameters__
        typevar_values = get_args(resolve_newtype_and_aliases(typ))
        assert len(typevars) == len(typevar_values)
        typ = origin_cls
        type_from_typevar.update(dict(zip(typevars, typevar_values)))
    elif (
        # Apply some heuristics for generic types. Should revisit this.
        hasattr(typ, "__parameters__") and hasattr(typ.__parameters__, "__len__")  # type: ignore
    ):
        typevars = typ.__parameters__  # type: ignore
        typevar_values = tuple(type_from_typevar_constraints(x) for x in typevars)
        assert len(typevars) == len(typevar_values)
        type_from_typevar.update(dict(zip(typevars, typevar_values)))

    if hasattr(typ, "__orig_bases__"):
        bases = getattr(typ, "__orig_bases__")
        for base in bases:
            origin_base = unwrap_origin_strip_extras(base)
            if origin_base is base or not hasattr(origin_base, "__parameters__"):
                continue
            typevars = origin_base.__parameters__
            typevar_values = get_args(base)
            type_from_typevar.update(dict(zip(typevars, typevar_values)))

    if len(annotations) == 0:
        return typ, type_from_typevar
    else:
        return (
            Annotated[(typ, *annotations)],  # type: ignore
            type_from_typevar,
        )


def get_type_hints_with_backported_syntax(
    obj: Callable[..., Any], include_extras: bool = False
) -> Dict[str, Any]:
    """Same as `typing.get_type_hints()`, but supports new union syntax (X | Y)
    and generics (list[str]) in older versions of Python."""
    try:
        out = get_type_hints(obj, include_extras=include_extras)

        # Workaround for:
        # - https://github.com/brentyi/tyro/issues/156
        # - https://github.com/python/cpython/issues/90353
        #
        # Which impacts Python 3.10 and earlier.
        #
        # It may be possible to remove this if this issue is resolved:
        # - https://github.com/python/typing_extensions/issues/310
        if sys.version_info < (3, 11):
            # If we see Optional[Annotated[T, ...]], we're going to flip to Annotated[Optional[T]]...
            #
            # It's unlikely but possible for this to have unintended side effects.
            for k, v in out.items():
                origin = get_origin(v)
                args = get_args(v)
                if (
                    origin is Union
                    and len(args) == 2
                    and (args[0] is NoneType or args[1] is NoneType)
                ):
                    non_none = args[1] if args[0] is NoneType else args[0]
                    if get_origin(non_none) is Annotated:
                        annotated_args = get_args(non_none)
                        out[k] = Annotated[  # type: ignore
                            (
                                Union.__getitem__((annotated_args[0], None)),  # type: ignore
                                *annotated_args[1:],
                            )
                        ]
        return out

    except TypeError as e:  # pragma: no cover
        # Resolve new type syntax using eval_type_backport.
        if hasattr(obj, "__annotations__"):
            try:
                from eval_type_backport import eval_type_backport

                # Get global namespace for functions.
                globalns = getattr(obj, "__globals__", None)

                # Get global namespace for classes.
                if globalns is None and hasattr(globalns, "__init__"):
                    globalns = getattr(getattr(obj, "__init__"), "__globals__", None)

                out = {
                    k: eval_type_backport(ForwardRef(v), globalns=globalns, localns={})
                    for k, v in getattr(obj, "__annotations__").items()
                }
                return out
            except ImportError:
                pass
        raise e



"""Multi-value Arguments

Arguments of both fixed and variable lengths can be annotated with standard
Python collection types. For Python 3.7 and 3.8, we can use either ``from
__future__ import annotations`` to support ``list[T]`` and ``tuple[T]``,
or the older :py:class:`typing.List` and :py:data:`typing.Tuple`.

Usage:

    # To print help:
    python ./03_multivalue.py --help

    # We can override arguments:
    python ./03_multivalue.py --source-paths ./data --dimensions 16 16
    python ./03_multivalue.py --source-paths ./data1 ./data2
"""

import pathlib
from dataclasses import dataclass
from pprint import pprint

import tyro


@dataclass
class Config:
    # Example of a variable-length tuple. `list[T]`, `set[T]`,
    # `dict[K, V]`, etc are supported as well.
    source_paths: tuple[pathlib.Path, ...]
    """This can be multiple!"""

    # Fixed-length tuples are also okay.
    dimensions: tuple[int, int] = (32, 32)
    """Height and width."""


if __name__ == "__main__":
    config = tyro.cli(Config)
    pprint(config)



from typing import List, Optional, Tuple

import pytest

import tyro


def test_positional():
    def main(
        x: int,
        y: int,
        /,
        # Note: it's generally a bad idea to have a mutable object (like a list) as a
        # default value. But it should still work.
        z: List[int] = [1, 2, 3],
    ) -> Tuple[int, int, int]:
        """main.

        Args:
            x: x
            y: y
            z: z

        Returns:
            Tuple[int, int, int]: Output.
        """
        return (x, y, z[0])

    assert tyro.cli(main, args="1 2 --z 3".split(" ")) == (1, 2, 3)
    with pytest.raises(SystemExit):
        assert tyro.cli(main, args="--x 1 --y 2 --z 3".split(" ")) == (1, 2, 3)


def test_nested_positional():
    class A:
        def __init__(self, a: int, hello_world: int, /, c: int):
            self.hello_world = hello_world

    def nest1(a: int, b: int, thing: A, /, c: int) -> A:
        return thing

    assert isinstance(tyro.cli(nest1, args="0 1 2 3 4 --c 4".split(" ")), A)
    assert tyro.cli(nest1, args="0 1 2 3 4 --c 4".split(" ")).hello_world == 3
    with pytest.raises(SystemExit):
        tyro.cli(nest1, args="0 1 2 3 4 4 --c 4".split(" "))


def test_nested_positional_alt():
    class B:
        def __init__(self, a: int, b: int, /, c: int):
            pass

    def nest2(a: int, b: int, /, thing: B, c: int):
        return thing

    assert isinstance(tyro.cli(nest2, args="0 1 2 3 --thing.c 4 --c 4".split(" ")), B)
    with pytest.raises(SystemExit):
        tyro.cli(nest2, args="0 1 2 3 4 --thing.c 4 --c 4".split(" "))


def test_positional_with_underscores():
    """Hyphen replacement works a bit different for positional arguments."""

    def main(a_multi_word_input: int, /) -> int:
        return a_multi_word_input

    assert tyro.cli(main, args=["5"]) == 5


def test_positional_booleans():
    """Make sure that flag behavior is disabled for positional booleans."""

    def main(
        flag1: bool,
        flag2: bool = True,
        flag3: bool = False,
        /,
    ) -> Tuple[bool, bool, bool]:
        return flag1, flag2, flag3

    assert tyro.cli(main, args=["True"]) == (True, True, False)
    assert tyro.cli(main, args=["True", "False"]) == (True, False, False)
    assert tyro.cli(main, args=["False", "False", "True"]) == (False, False, True)

    with pytest.raises(SystemExit):
        tyro.cli(main, args=["hmm"])
    with pytest.raises(SystemExit):
        tyro.cli(main, args=["true"])
    with pytest.raises(SystemExit):
        tyro.cli(main, args=["True", "false"])


def test_optional_list():
    def main(a: Optional[List[int]], /) -> Optional[List[int]]:
        return a

    assert tyro.cli(main, args=["None"]) is None
    assert tyro.cli(main, args=["1", "2"]) == [1, 2]
    assert tyro.cli(main, args=[]) == []
    with pytest.raises(SystemExit):
        tyro.cli(main, args=["hm"])


def test_optional_list_with_default():
    def main(a: Optional[List[int]] = None, /) -> Optional[List[int]]:
        return a

    assert tyro.cli(main, args=["None"]) is None
    assert tyro.cli(main, args=["5", "5"]) == [5, 5]
    with pytest.raises(SystemExit):
        tyro.cli(main, args=["None", "5"])


def test_positional_tuple():
    def main(x: Tuple[int, int], y: Tuple[str, str], /):
        return x, y

    assert tyro.cli(main, args="1 2 3 4".split(" ")) == ((1, 2), ("3", "4"))



"""Nested Dataclasses

Structures (typically :py:func:`dataclasses.dataclass`) can be nested to build hierarchical configuration
objects. This helps with modularity and grouping in larger projects.

Usage:

    python ./01_nesting.py --help
    python ./01_nesting.py --opt.learning-rate 1e-3
    python ./01_nesting.py --seed 4
"""

import dataclasses

import tyro


@dataclasses.dataclass
class OptimizerConfig:
    learning_rate: float = 3e-4
    weight_decay: float = 1e-2


@dataclasses.dataclass
class Config:
    # Optimizer options.
    opt: OptimizerConfig

    # Random seed.
    seed: int = 0


if __name__ == "__main__":
    config = tyro.cli(Config)
    print(dataclasses.asdict(config))



"""Interface for generating `argparse.ArgumentParser()` definitions from callables."""

from __future__ import annotations

import dataclasses
from typing import (
    Any,
    Callable,
    Dict,
    List,
    Optional,
    Set,
    Tuple,
    Type,
    TypeVar,
    Union,
    cast,
)

from typing_extensions import Annotated, get_args, get_origin

from tyro.constructors._registry import ConstructorRegistry
from tyro.constructors._struct_spec import UnsupportedStructTypeMessage

from . import _argparse as argparse
from . import (
    _argparse_formatter,
    _arguments,
    _docstrings,
    _fields,
    _resolver,
    _singleton,
    _strings,
    _subcommand_matching,
)
from ._typing import TypeForm
from .conf import _confstruct, _markers
from .constructors._primitive_spec import (
    PrimitiveConstructorSpec,
    PrimitiveTypeInfo,
    UnsupportedTypeAnnotationError,
)

T = TypeVar("T")


@dataclasses.dataclass(frozen=True)
class ParserSpecification:
    """Each parser contains a list of arguments and optionally some subparsers."""

    f: Callable
    description: str
    args: List[_arguments.ArgumentDefinition]
    field_list: List[_fields.FieldDefinition]
    child_from_prefix: Dict[str, ParserSpecification]
    helptext_from_intern_prefixed_field_name: Dict[str, Optional[str]]

    # We have two mechanics for tracking subparser groups:
    # - A single subparser group, which is what gets added in the tree structure built
    # by the argparse parser.
    subparsers: Optional[SubparsersSpecification]
    # - A set of subparser groups, which reflect the tree structure built by the
    # hierarchy of a nested config structure.
    subparsers_from_intern_prefix: Dict[str, SubparsersSpecification]
    intern_prefix: str
    extern_prefix: str
    has_required_args: bool
    consolidate_subcommand_args: bool

    @staticmethod
    def from_callable_or_type(
        f: Callable[..., T],
        markers: Set[_markers._Marker],
        description: Optional[str],
        parent_classes: Set[Type[Any]],
        default_instance: Union[
            T, _singleton.PropagatingMissingType, _singleton.NonpropagatingMissingType
        ],
        intern_prefix: str,
        extern_prefix: str,
        subcommand_prefix: str = "",
        support_single_arg_types: bool = False,
    ) -> ParserSpecification:
        """Create a parser definition from a callable or type."""

        # Consolidate subcommand types.
        markers = markers | set(_resolver.unwrap_annotated(f, _markers._Marker)[1])
        consolidate_subcommand_args = _markers.ConsolidateSubcommandArgs in markers

        # Resolve the type of `f`, generate a field list.
        with _fields.FieldDefinition.marker_context(tuple(markers)):
            out = _fields.field_list_from_type_or_callable(
                f=f,
                default_instance=default_instance,
                support_single_arg_types=support_single_arg_types,
            )
            assert not isinstance(out, UnsupportedStructTypeMessage)
            f, field_list = out

        # Cycle detection.
        #
        # 'parent' here refers to in the nesting hierarchy, not the superclass.
        if f in parent_classes and f is not dict:
            raise UnsupportedTypeAnnotationError(
                f"Found a cyclic dependency with type {f}."
            )

        # TODO: we are abusing the (minor) distinctions between types, classes, and
        # callables throughout the code. This is mostly for legacy reasons, could be
        # cleaned up.
        parent_classes = parent_classes | {cast(Type, f)}

        has_required_args = False
        args = []
        helptext_from_intern_prefixed_field_name: Dict[str, Optional[str]] = {}

        child_from_prefix: Dict[str, ParserSpecification] = {}

        subparsers = None
        subparsers_from_prefix = {}

        for field in field_list:
            field_out = handle_field(
                field,
                parent_classes=parent_classes,
                intern_prefix=intern_prefix,
                extern_prefix=extern_prefix,
                subcommand_prefix=subcommand_prefix,
            )
            if isinstance(field_out, _arguments.ArgumentDefinition):
                # Handle single arguments.
                args.append(field_out)
                if field_out.lowered.required:
                    has_required_args = True
            elif isinstance(field_out, SubparsersSpecification):
                # Handle subparsers.
                subparsers_from_prefix[field_out.intern_prefix] = field_out
                subparsers = add_subparsers_to_leaves(subparsers, field_out)
            elif isinstance(field_out, ParserSpecification):
                # Handle nested parsers.
                nested_parser = field_out
                child_from_prefix[field_out.intern_prefix] = nested_parser

                if nested_parser.has_required_args:
                    has_required_args = True

                # Include nested subparsers.
                if nested_parser.subparsers is not None:
                    subparsers_from_prefix.update(
                        nested_parser.subparsers_from_intern_prefix
                    )
                    subparsers = add_subparsers_to_leaves(
                        subparsers, nested_parser.subparsers
                    )

                # Helptext for this field; used as description for grouping arguments.
                class_field_name = _strings.make_field_name(
                    [intern_prefix, field.intern_name]
                )
                if field.helptext is not None:
                    helptext_from_intern_prefixed_field_name[class_field_name] = (
                        field.helptext
                    )
                else:
                    helptext_from_intern_prefixed_field_name[class_field_name] = (
                        _docstrings.get_callable_description(nested_parser.f)
                    )

                # If arguments are in an optional group, it indicates that the default_instance
                # will be used if none of the arguments are passed in.
                if (
                    len(nested_parser.args) >= 1
                    and _markers._OPTIONAL_GROUP in nested_parser.args[0].field.markers
                ):
                    current_helptext = helptext_from_intern_prefixed_field_name[
                        class_field_name
                    ]
                    helptext_from_intern_prefixed_field_name[class_field_name] = (
                        ("" if current_helptext is None else current_helptext + "\n\n")
                        + "Default: "
                        + str(field.default)
                    )

        return ParserSpecification(
            f=f,
            description=_strings.remove_single_line_breaks(
                description
                if description is not None
                else _docstrings.get_callable_description(f)
            ),
            args=args,
            field_list=field_list,
            child_from_prefix=child_from_prefix,
            helptext_from_intern_prefixed_field_name=helptext_from_intern_prefixed_field_name,
            subparsers=subparsers,
            subparsers_from_intern_prefix=subparsers_from_prefix,
            intern_prefix=intern_prefix,
            extern_prefix=extern_prefix,
            has_required_args=has_required_args,
            consolidate_subcommand_args=consolidate_subcommand_args,
        )

    def apply(
        self, parser: argparse.ArgumentParser
    ) -> Tuple[argparse.ArgumentParser, ...]:
        """Create defined arguments and subparsers."""

        # Generate helptext.
        parser.description = self.description

        # Create subparser tree.
        subparser_group = None
        if self.subparsers is not None:
            leaves = self.subparsers.apply(parser)
            subparser_group = parser._action_groups.pop()
        else:
            leaves = (parser,)

        # Depending on whether we want to consolidate subcommand args, we can either
        # apply arguments to the intermediate parser or only on the leaves.
        if self.consolidate_subcommand_args:
            for leaf in leaves:
                self.apply_args(leaf)
        else:
            self.apply_args(parser)

        if subparser_group is not None:
            parser._action_groups.append(subparser_group)

        # Break some API boundaries to rename the "optional arguments" => "options".
        assert parser._action_groups[1].title in (
            # python <= 3.9
            "optional arguments",
            # python >= 3.10
            "options",
        )
        parser._action_groups[1].title = "options"

        return leaves

    def apply_args(
        self,
        parser: argparse.ArgumentParser,
        parent: Optional[ParserSpecification] = None,
    ) -> None:
        """Create defined arguments and subparsers."""

        # Make argument groups.
        def format_group_name(group_name: str) -> str:
            return (group_name + " options").strip()

        def group_name_from_arg(arg: _arguments.ArgumentDefinition) -> str:
            prefix = arg.lowered.name_or_flag
            if prefix.startswith("--"):
                prefix = prefix[2:]
            if "." in prefix:
                prefix = prefix.rpartition(".")[0]
            else:
                prefix = ""
            return prefix

        group_from_group_name: Dict[str, argparse._ArgumentGroup] = {
            "": parser._action_groups[1],
            **{
                cast(str, group.title).partition(" ")[0]: group
                for group in parser._action_groups[2:]
            },
        }
        positional_group = parser._action_groups[0]
        assert positional_group.title == "positional arguments"

        # Add each argument group. Groups with only suppressed arguments won't
        # be added.
        for arg in self.args:
            group_name = group_name_from_arg(arg)
            if (
                arg.lowered.help is not argparse.SUPPRESS
                and group_name not in group_from_group_name
            ):
                description = (
                    parent.helptext_from_intern_prefixed_field_name.get(
                        arg.intern_prefix
                    )
                    if parent is not None
                    else None
                )
                group_from_group_name[group_name] = parser.add_argument_group(
                    format_group_name(group_name),
                    description=description,
                )

            # Add each argument.
            if arg.field.is_positional():
                arg.add_argument(positional_group)
                continue

            if group_name in group_from_group_name:
                arg.add_argument(group_from_group_name[group_name])
            else:
                # Suppressed argument: still need to add them, but they won't show up in
                # the helptext so it doesn't matter which group.
                assert arg.lowered.help is argparse.SUPPRESS
                arg.add_argument(group_from_group_name[""])

        for child in self.child_from_prefix.values():
            child.apply_args(parser, parent=self)


def handle_field(
    field: _fields.FieldDefinition,
    parent_classes: Set[Type[Any]],
    intern_prefix: str,
    extern_prefix: str,
    subcommand_prefix: str,
) -> Union[
    _arguments.ArgumentDefinition,
    ParserSpecification,
    SubparsersSpecification,
]:
    """Determine what to do with a single field definition."""

    registry = ConstructorRegistry._get_active_registry()

    # Force primitive if (1) the field is annotated with a primitive constructor spec, or (2) if
    force_primitive = len(
        _resolver.unwrap_annotated(field.type, PrimitiveConstructorSpec)[1]
    ) > 0 or (
        len(registry._custom_primitive_rules) > 0
        and registry.get_primitive_spec(
            PrimitiveTypeInfo.make(field.type, field.markers), rule_mode="custom"
        )
        is not None
    )
    if (
        not force_primitive
        and _markers.Fixed not in field.markers
        and _markers.Suppress not in field.markers
    ):
        # (1) Handle Unions over callables; these result in subparsers.
        subparsers_attempt = SubparsersSpecification.from_field(
            field,
            parent_classes=parent_classes,
            intern_prefix=_strings.make_field_name([intern_prefix, field.intern_name]),
            extern_prefix=_strings.make_field_name([extern_prefix, field.extern_name]),
        )
        if subparsers_attempt is not None:
            if (
                not subparsers_attempt.required
                and _markers.AvoidSubcommands in field.markers
            ):
                # Don't make a subparser.
                field = field.with_new_type_stripped(type(field.default))
            else:
                return subparsers_attempt

        # (2) Handle nested callables.
        if force_primitive == "struct" or _fields.is_struct_type(
            field.type_stripped, field.default
        ):
            field = field.with_new_type_stripped(
                _resolver.narrow_subtypes(
                    field.type_stripped,
                    field.default,
                ),
            )
            return ParserSpecification.from_callable_or_type(
                field.type_stripped,
                markers=field.markers,
                description=None,
                parent_classes=parent_classes,
                default_instance=field.default,
                intern_prefix=_strings.make_field_name(
                    [intern_prefix, field.intern_name]
                ),
                extern_prefix=(
                    _strings.make_field_name([extern_prefix, field.extern_name])
                    if field.argconf.prefix_name in (True, None)
                    else field.extern_name
                ),
                subcommand_prefix=subcommand_prefix,
                support_single_arg_types=False,
            )

    # (3) Handle primitive or fixed types. These produce a single argument!
    return _arguments.ArgumentDefinition(
        intern_prefix=intern_prefix,
        extern_prefix=extern_prefix,
        subcommand_prefix=subcommand_prefix,
        field=field,
    )


@dataclasses.dataclass(frozen=True)
class SubparsersSpecification:
    """Structure for defining subparsers. Each subparser is a parser with a name."""

    name: str
    description: Optional[str]
    parser_from_name: Dict[str, ParserSpecification]
    intern_prefix: str
    required: bool
    default_instance: Any
    options: Tuple[Union[TypeForm[Any], Callable], ...]

    @staticmethod
    def from_field(
        field: _fields.FieldDefinition,
        parent_classes: Set[Type[Any]],
        intern_prefix: str,
        extern_prefix: str,
    ) -> Optional[SubparsersSpecification]:
        # Union of classes should create subparsers.
        typ = _resolver.unwrap_annotated(field.type_stripped)
        if get_origin(typ) not in (Union, _resolver.UnionType):
            return None

        # We don't use sets here to retain order of subcommands.
        options: List[Union[type, Callable]]
        options = [typ for typ in get_args(typ)]
        options = [
            (
                # Cast seems unnecessary but needed in mypy... (1.4.1)
                cast(Callable, none_proxy) if o is type(None) else o
            )
            for o in options
        ]

        # If specified, swap types using tyro.conf.subcommand(constructor=...).
        for i, option in enumerate(options):
            _, found_subcommand_configs = _resolver.unwrap_annotated(
                option, _confstruct._SubcommandConfig
            )
            if (
                len(found_subcommand_configs) > 0
                and found_subcommand_configs[0].constructor_factory is not None
            ):
                options[i] = Annotated[  # type: ignore
                    (
                        found_subcommand_configs[0].constructor_factory(),
                        *_resolver.unwrap_annotated(option, "all")[1],
                    )
                ]

        # Exit if we don't contain any nested types.
        if not any(
            [
                o is not none_proxy
                and _fields.is_struct_type(cast(type, o), _singleton.MISSING_NONPROP)
                for o in options
            ]
        ):
            return None

        # Get subcommand configurations from `tyro.conf.subcommand()`.
        subcommand_config_from_name: Dict[str, _confstruct._SubcommandConfig] = {}
        subcommand_type_from_name: Dict[str, type] = {}
        for option in options:
            subcommand_name = _strings.subparser_name_from_type(
                (
                    ""
                    if _markers.OmitSubcommandPrefixes in field.markers
                    else extern_prefix
                ),
                type(None) if option is none_proxy else cast(type, option),
            )
            option_unwrapped, found_subcommand_configs = _resolver.unwrap_annotated(
                option, _confstruct._SubcommandConfig
            )
            if len(found_subcommand_configs) != 0:
                # Explicitly annotated default.
                assert len(found_subcommand_configs) == 1, (
                    f"Expected only one subcommand config, but {subcommand_name} has"
                    f" {len(found_subcommand_configs)}."
                )
                subcommand_config_from_name[subcommand_name] = found_subcommand_configs[
                    0
                ]
            subcommand_type_from_name[subcommand_name] = cast(type, option)

        # If a field default is provided, try to find a matching subcommand name.
        if (
            field.default is None
            or field.default in _singleton.MISSING_AND_MISSING_NONPROP
        ):
            default_name = None
        else:
            default_name = _subcommand_matching.match_subcommand(
                field.default,
                subcommand_config_from_name,
                subcommand_type_from_name,
            )

            assert default_name is not None, (
                f"`{extern_prefix}` was provided a default value of type"
                f" {type(field.default)} but no matching subcommand was found. A"
                " type may be missing in the Union type declaration for"
                f" `{extern_prefix}`, which currently expects {options}. "
                "The types may also be too complex for tyro's subcommand matcher; support "
                "is particularly limited for custom generic types."
            )

        # Add subcommands for each option.
        parser_from_name: Dict[str, ParserSpecification] = {}
        for option in options:
            subcommand_name = _strings.subparser_name_from_type(
                (
                    ""
                    if _markers.OmitSubcommandPrefixes in field.markers
                    else extern_prefix
                ),
                type(None) if option is none_proxy else cast(type, option),
            )

            # Get a subcommand config: either pulled from the type annotations or the
            # field default.
            if subcommand_name in subcommand_config_from_name:
                subcommand_config = subcommand_config_from_name[subcommand_name]
            else:
                subcommand_config = _confstruct._SubcommandConfig(
                    "unused",
                    description=None,
                    default=_singleton.MISSING_NONPROP,
                    prefix_name=True,
                    constructor_factory=None,
                )

            # If names match, borrow subcommand default from field default.
            if default_name == subcommand_name and (
                field.is_default_from_default_instance
                or subcommand_config.default in _singleton.MISSING_AND_MISSING_NONPROP
            ):
                subcommand_config = dataclasses.replace(
                    subcommand_config, default=field.default
                )

            # Strip the subcommand config from the option type.
            # Relevant: https://github.com/brentyi/tyro/pull/117
            option_origin, annotations = _resolver.unwrap_annotated(option, "all")
            annotations = tuple(
                a
                for a in annotations
                if not isinstance(a, _confstruct._SubcommandConfig)
            )
            if len(annotations) == 0:
                option = option_origin
            else:
                option = Annotated[(option_origin,) + annotations]  # type: ignore

            with _fields.FieldDefinition.marker_context(tuple(field.markers)):
                subparser = ParserSpecification.from_callable_or_type(
                    option,  # type: ignore
                    markers=field.markers,
                    description=subcommand_config.description,
                    parent_classes=parent_classes,
                    default_instance=subcommand_config.default,
                    intern_prefix=intern_prefix,
                    extern_prefix=extern_prefix,
                    subcommand_prefix=intern_prefix,
                    support_single_arg_types=True,
                )

            # Apply prefix to helptext in nested classes in subparsers.
            subparser = dataclasses.replace(
                subparser,
                helptext_from_intern_prefixed_field_name={
                    _strings.make_field_name([intern_prefix, k]): v
                    for k, v in subparser.helptext_from_intern_prefixed_field_name.items()
                },
            )
            parser_from_name[subcommand_name] = subparser

        # Required if a default is missing.
        required = field.default in _fields.MISSING_AND_MISSING_NONPROP

        # Required if a default is passed in, but the default value has missing
        # parameters.
        if default_name is not None:
            default_parser = parser_from_name[default_name]
            if any(map(lambda arg: arg.lowered.required, default_parser.args)):
                required = True
            if (
                default_parser.subparsers is not None
                and default_parser.subparsers.required
            ):
                required = True

        # Required if all args are pushed to the final subcommand.
        if _markers.ConsolidateSubcommandArgs in field.markers:
            required = True

        # Make description.
        description_parts = []
        if field.helptext is not None:
            description_parts.append(field.helptext)
        if not required and field.default not in _fields.MISSING_AND_MISSING_NONPROP:
            description_parts.append(f" (default: {default_name})")
        description = (
            # We use `None` instead of an empty string to prevent a line break from
            # being created where the description would be.
            " ".join(description_parts) if len(description_parts) > 0 else None
        )

        return SubparsersSpecification(
            name=field.intern_name,
            # If we wanted, we could add information about the default instance
            # automatically, as is done for normal fields. But for now we just rely on
            # the user to include it in the docstring.
            description=description,
            parser_from_name=parser_from_name,
            intern_prefix=intern_prefix,
            required=required,
            default_instance=field.default,
            options=tuple(options),
        )

    def apply(
        self, parent_parser: argparse.ArgumentParser
    ) -> Tuple[argparse.ArgumentParser, ...]:
        title = "subcommands"
        metavar = "{" + ",".join(self.parser_from_name.keys()) + "}"
        if not self.required:
            title = "optional " + title
            metavar = f"[{metavar}]"

        # Add subparsers to every node in previous level of the tree.
        argparse_subparsers = parent_parser.add_subparsers(
            dest=_strings.make_subparser_dest(self.intern_prefix),
            description=self.description,
            required=self.required,
            title=title,
            metavar=metavar,
        )

        subparser_tree_leaves: List[argparse.ArgumentParser] = []
        for name, subparser_def in self.parser_from_name.items():
            helptext = subparser_def.description.replace("%", "%%")
            if len(helptext) > 0:
                # TODO: calling a private function here.
                helptext = _arguments._rich_tag_if_enabled(helptext.strip(), "helptext")

            subparser = argparse_subparsers.add_parser(
                name,
                formatter_class=_argparse_formatter.TyroArgparseHelpFormatter,
                help=helptext,
                allow_abbrev=False,
            )

            # Attributes used for error message generation.
            assert isinstance(subparser, _argparse_formatter.TyroArgumentParser)
            assert isinstance(parent_parser, _argparse_formatter.TyroArgumentParser)
            subparser._parsing_known_args = parent_parser._parsing_known_args
            subparser._parser_specification = parent_parser._parser_specification
            subparser._console_outputs = parent_parser._console_outputs
            subparser._args = parent_parser._args

            subparser_tree_leaves.extend(subparser_def.apply(subparser))

        return tuple(subparser_tree_leaves)


def add_subparsers_to_leaves(
    root: Optional[SubparsersSpecification], leaf: SubparsersSpecification
) -> SubparsersSpecification:
    if root is None:
        return leaf

    new_parsers_from_name = {}
    for name, parser in root.parser_from_name.items():
        new_parsers_from_name[name] = dataclasses.replace(
            parser,
            subparsers=add_subparsers_to_leaves(parser.subparsers, leaf),
        )
    return dataclasses.replace(
        root,
        parser_from_name=new_parsers_from_name,
        required=root.required or leaf.required,
    )


def none_proxy() -> None:
    return None



import contextlib
import dataclasses
import io
from typing import Dict, List, Tuple, TypeVar, Union

import pytest
from typing_extensions import Literal

import tyro
from tyro.constructors import UnsupportedTypeAnnotationError


def test_ambiguous_collection_0() -> None:
    @dataclasses.dataclass
    class A:
        x: Tuple[Tuple[int, ...], ...]

    with pytest.raises(UnsupportedTypeAnnotationError):
        tyro.cli(A, args=["--x", "0", "1"])


def test_ambiguous_collection_1() -> None:
    @dataclasses.dataclass
    class A:
        x: List[List[int]]

    with pytest.raises(UnsupportedTypeAnnotationError):
        tyro.cli(A, args=["--x", "0", "1"])


def test_ambiguous_collection_2() -> None:
    def main(x: Tuple[List[str], List[str]]) -> None:
        pass

    with pytest.raises(UnsupportedTypeAnnotationError) as e:
        tyro.cli(main, args=["--help"])
    assert "Unsupported type annotation for the field x" in e.value.args[0]


def test_ambiguous_collection_3() -> None:
    def main(x: List[Union[Tuple[int, int], Tuple[int, int, int]]]) -> None:
        pass

    with pytest.raises(UnsupportedTypeAnnotationError) as e:
        tyro.cli(main, args=["--help"])
    assert "Unsupported type annotation for the field x" in e.value.args[0]


def test_ambiguous_collection_4() -> None:
    X = List[Union[Tuple[int, int], Tuple[int, int, int]]]
    with pytest.raises(UnsupportedTypeAnnotationError) as e:
        tyro.cli(X, args=["--help"])
    assert "Unsupported type annotation for the field" not in e.value.args[0]


def test_ambiguous_collection_5() -> None:
    @dataclasses.dataclass
    class A:
        x: Dict[List[int], str]

    with pytest.raises(UnsupportedTypeAnnotationError):
        tyro.cli(A, args=["--x", "0", "1"])


def test_ambiguous_collection_6() -> None:
    @dataclasses.dataclass
    class A:
        x: Dict[str, List[int]]

    with pytest.raises(UnsupportedTypeAnnotationError):
        tyro.cli(A, args=["--x", "0", "1"])


# Must be global.
@dataclasses.dataclass
class _CycleDataclass:
    x: "_CycleDataclass"


def test_cycle() -> None:
    with pytest.raises(UnsupportedTypeAnnotationError):
        tyro.cli(_CycleDataclass, args=[])


def test_uncallable_annotation() -> None:
    def main(arg: 5) -> None:  # type: ignore
        pass

    with pytest.raises(UnsupportedTypeAnnotationError):
        tyro.cli(main, args=[])


def test_nested_annotation() -> None:
    @dataclasses.dataclass
    class OneIntArg:
        x: int

    def main(arg: List[OneIntArg]) -> List[OneIntArg]:  # type: ignore
        return arg

    with pytest.raises(UnsupportedTypeAnnotationError):
        tyro.cli(main, args=[])

    @dataclasses.dataclass
    class OneStringArg:
        x: str

    def main(arg: List[OneStringArg]) -> List[OneStringArg]:  # type: ignore
        return arg

    with pytest.raises(UnsupportedTypeAnnotationError):
        tyro.cli(main, args=["--arg", "0", "1", "2"])

    @dataclasses.dataclass
    class TwoStringArg:
        x: str
        y: str

    def main2(arg: List[TwoStringArg]) -> None:
        pass

    with pytest.raises(UnsupportedTypeAnnotationError):
        tyro.cli(main2, args=[])


def test_missing_annotation_1() -> None:
    def main(a, b) -> None:
        pass

    with pytest.raises(UnsupportedTypeAnnotationError):
        tyro.cli(main, args=["--help"])


def test_missing_annotation_2() -> None:
    def main(*, a) -> None:
        pass

    with pytest.raises(UnsupportedTypeAnnotationError):
        tyro.cli(main, args=["--help"])


def test_tuple_needs_default() -> None:
    def main(arg: tuple) -> None:  # type: ignore
        pass

    # This formerly raised an error, but now defaults to Tuple[str, ...].
    #  with pytest.raises(UnsupportedTypeAnnotationError):
    with pytest.raises(SystemExit):
        tyro.cli(main, args=["--help"])


def test_unbound_typevar() -> None:
    T = TypeVar("T")

    def main(arg: T) -> None:  # type: ignore
        pass

    with pytest.raises(UnsupportedTypeAnnotationError):
        tyro.cli(main, args=["--help"])


def test_missing_default_fixed() -> None:
    def main(value: tyro.conf.SuppressFixed[tyro.conf.Fixed[int]]) -> int:
        return value

    with pytest.raises(UnsupportedTypeAnnotationError):
        tyro.cli(main, args=["--help"])


def test_missing_default_suppressed() -> None:
    def main(value: tyro.conf.Suppress[int]) -> int:
        return value

    with pytest.raises(UnsupportedTypeAnnotationError):
        tyro.cli(main, args=["--help"])


def test_ambiguous_sequence() -> None:
    def main(value: list) -> None:
        return None

    # This formerly raised an error, but now defaults to List[str].
    #  with pytest.raises(UnsupportedTypeAnnotationError):
    with pytest.raises(SystemExit):
        tyro.cli(main, args=["--help"])


def test_similar_arguments_basic() -> None:
    @dataclasses.dataclass
    class RewardConfig:
        track: bool

    @dataclasses.dataclass
    class Class:
        reward: RewardConfig

    target = io.StringIO()
    with pytest.raises(SystemExit), contextlib.redirect_stderr(target):
        tyro.cli(Class, args="--reward.trac".split(" "))

    error = target.getvalue()
    assert "Unrecognized option" in error
    assert "Perhaps you meant:" in error

    assert error.count("--reward.track") == 1
    assert error.count("--help") == 1


def test_suppress_console_outputs() -> None:
    @dataclasses.dataclass
    class RewardConfig:
        track: bool

    @dataclasses.dataclass
    class Class:
        reward: RewardConfig

    target = io.StringIO()
    with pytest.raises(SystemExit), contextlib.redirect_stderr(target):
        tyro.cli(Class, args="--reward.trac".split(" "), console_outputs=False)

    error = target.getvalue()
    assert error == ""


def test_suppress_console_outputs_fromdict() -> None:
    def foo(track: bool) -> None:
        print(track)

    def bar(track: bool) -> None:
        print(track)

    target = io.StringIO()
    with pytest.raises(SystemExit), contextlib.redirect_stderr(target):
        tyro.extras.subcommand_cli_from_dict(
            {"foo": foo, "bar": bar},
            args="foo --reward.trac".split(" "),
            console_outputs=False,
        )

    error = target.getvalue()
    assert error == ""


def test_similar_arguments_subcommands() -> None:
    @dataclasses.dataclass
    class RewardConfig:
        track: bool

    @dataclasses.dataclass
    class ClassA:
        reward: RewardConfig

    @dataclasses.dataclass
    class ClassB:
        reward: RewardConfig

    target = io.StringIO()
    with pytest.raises(SystemExit), contextlib.redirect_stderr(target):
        tyro.cli(Union[ClassA, ClassB], args="--reward.trac".split(" "))  # type: ignore

    error = target.getvalue()
    assert "Unrecognized option" in error
    assert "Perhaps you meant:" in error
    assert error.count("--reward.track") == 1
    assert error.count("--help") == 3


def test_similar_arguments_subcommands_multiple() -> None:
    @dataclasses.dataclass
    class RewardConfig:
        track: bool
        trace: int

    @dataclasses.dataclass
    class ClassA:
        reward: RewardConfig

    @dataclasses.dataclass
    class ClassB:
        reward: RewardConfig

    target = io.StringIO()
    with pytest.raises(SystemExit), contextlib.redirect_stderr(target):
        tyro.cli(Union[ClassA, ClassB], args="--fjdkslaj --reward.trac".split(" "))  # type: ignore

    error = target.getvalue()
    assert "Unrecognized option" in error
    assert "Arguments similar to --reward.trac" in error
    assert error.count("--reward.track {True,False}") == 1
    assert error.count("--reward.trace INT") == 1
    assert error.count("--help") == 5


def test_similar_arguments_subcommands_multiple_contains_match() -> None:
    @dataclasses.dataclass
    class RewardConfigA:
        track: bool
        trace: int

    @dataclasses.dataclass
    class RewardConfigB: ...

    @dataclasses.dataclass
    class ClassA:
        reward: RewardConfigA

    @dataclasses.dataclass
    class ClassB:
        reward: RewardConfigB

    target = io.StringIO()
    with pytest.raises(SystemExit), contextlib.redirect_stderr(target):
        tyro.cli(
            Union[ClassA, ClassB],
            args="class-b --reward.track True --reward.trace 7".split(" "),
        )  # type: ignore

    error = target.getvalue()
    assert "Unrecognized or misplaced" in error
    assert (
        "(applied to " in error and "class-b)" in error
    )  # (applied to {root prog} class-b)
    assert "Arguments similar to" in error
    assert error.count("--reward.track {True,False}") == 2
    assert error.count("--reward.trace INT") == 2
    assert error.count("--help") == 5  # 2 subcommands * 2 arguments + usage hint.


def test_similar_arguments_subcommands_multiple_contains_match_alt() -> None:
    @dataclasses.dataclass
    class RewardConfig:
        track: bool
        trace: int

    @dataclasses.dataclass
    class ClassA:
        reward: RewardConfig

    @dataclasses.dataclass
    class ClassB:
        reward: RewardConfig

    target = io.StringIO()
    with pytest.raises(SystemExit), contextlib.redirect_stderr(target):
        tyro.cli(Union[ClassA, ClassB], args="--track".split(" "))  # type: ignore

    error = target.getvalue()
    assert "Unrecognized option" in error
    assert "Perhaps you meant:" in error
    assert error.count("--reward.track {True,False}") == 1
    assert (
        error.count("--help") == 3
    )  # Should show two possible subcommands + usage hint.


def test_similar_arguments_subcommands_overflow_different() -> None:
    @dataclasses.dataclass
    class RewardConfig:
        track0: bool
        track1: bool
        track2: bool
        track3: bool
        track4: bool
        track5: bool
        track6: bool
        track7: bool
        track8: bool
        track9: bool
        track10: bool
        track11: bool
        track12: bool
        track13: bool
        track14: bool
        track15: bool
        track16: bool

    @dataclasses.dataclass
    class ClassA:
        reward: RewardConfig

    @dataclasses.dataclass
    class ClassB:
        reward: RewardConfig

    target = io.StringIO()
    with pytest.raises(SystemExit), contextlib.redirect_stderr(target):
        tyro.cli(Union[ClassA, ClassB], args="--track".split(" "))  # type: ignore

    error = target.getvalue()
    assert "Unrecognized option" in error
    assert "Perhaps you meant:" in error
    assert error.count("--reward.track") == 10
    assert "[...]" not in error
    assert error.count("--help") == 21

    target = io.StringIO()
    with pytest.raises(SystemExit), contextlib.redirect_stderr(target):
        # --tracked is intentionally between 0.8 ~ 0.9 similarity to track{i} for test
        # coverage.
        tyro.cli(RewardConfig, args="--tracked".split(" "))  # type: ignore

    # Usage print should be clipped.
    error = target.getvalue()
    assert "For full helptext, run" in error


def test_similar_arguments_subcommands_overflow_same() -> None:
    @dataclasses.dataclass
    class RewardConfig:
        track: bool

    @dataclasses.dataclass
    class ClassA:
        reward: RewardConfig

    @dataclasses.dataclass
    class ClassB:
        reward: RewardConfig

    @dataclasses.dataclass
    class ClassC:
        reward: RewardConfig

    @dataclasses.dataclass
    class ClassD:
        reward: RewardConfig

    @dataclasses.dataclass
    class ClassE:
        reward: RewardConfig

    @dataclasses.dataclass
    class ClassF:
        reward: RewardConfig

    @dataclasses.dataclass
    class ClassG:
        reward: RewardConfig

    @dataclasses.dataclass
    class ClassH:
        reward: RewardConfig

    @dataclasses.dataclass
    class ClassI:
        reward: RewardConfig

    target = io.StringIO()
    with pytest.raises(SystemExit), contextlib.redirect_stderr(target):
        tyro.cli(  # type: ignore
            Union[
                ClassA, ClassB, ClassC, ClassD, ClassE, ClassF, ClassG, ClassH, ClassI
            ],
            args="--track".split(" "),
        )

    error = target.getvalue()
    assert "Unrecognized option" in error
    assert "Perhaps you meant:" in error
    assert error.count("--reward.track") == 1
    assert "[...]" in error
    assert error.count("--help") == 5


def test_similar_arguments_subcommands_overflow_same_startswith_multiple() -> None:
    @dataclasses.dataclass
    class RewardConfig:
        track: bool

    @dataclasses.dataclass
    class ClassA:
        reward: RewardConfig

    @dataclasses.dataclass
    class ClassB:
        reward: RewardConfig

    @dataclasses.dataclass
    class ClassC:
        reward: RewardConfig

    @dataclasses.dataclass
    class ClassD:
        reward: RewardConfig

    @dataclasses.dataclass
    class ClassE:
        reward: RewardConfig
        rewarde: tyro.conf.Suppress[int] = 10

    @dataclasses.dataclass
    class ClassF:
        reward: RewardConfig

    @dataclasses.dataclass
    class ClassG:
        reward: RewardConfig

    @dataclasses.dataclass
    class ClassH:
        reward: RewardConfig

    @dataclasses.dataclass
    class ClassI:
        reward: RewardConfig

    target = io.StringIO()
    with pytest.raises(SystemExit), contextlib.redirect_stderr(target):
        tyro.cli(  # type: ignore
            Union[
                ClassA, ClassB, ClassC, ClassD, ClassE, ClassF, ClassG, ClassH, ClassI
            ],
            args="--track --ffff".split(" "),
        )

    error = target.getvalue()
    assert "Unrecognized option" in error
    assert "Arguments similar to --track" in error
    assert error.count("--rewar") == 1
    assert "rewarde" not in error
    assert "[...]" in error
    assert error.count("--help") == 5


def test_similar_flag() -> None:
    @dataclasses.dataclass
    class Args:
        flag: bool = False

    target = io.StringIO()
    with pytest.raises(SystemExit), contextlib.redirect_stderr(target):
        tyro.cli(
            Args,
            args="--lag".split(" "),
        )

    error = target.getvalue()

    # We don't print usage text anymore.
    assert error.count("--flag | --no-flag") == 0

    # Printed in the similar argument list.
    assert error.count("--flag, --no-flag") == 1


def test_value_error() -> None:
    """https://github.com/brentyi/tyro/issues/86"""

    def main() -> None:
        raise ValueError("This shouldn't be caught by tyro")

    with pytest.raises(ValueError):
        tyro.cli(main, args=[])


def test_value_error_subcommand() -> None:
    """https://github.com/brentyi/tyro/issues/86"""

    def main1() -> None:
        raise ValueError("This shouldn't be caught by tyro")

    def main2() -> None:
        raise ValueError("This shouldn't be caught by tyro")

    with pytest.raises(ValueError):
        tyro.extras.subcommand_cli_from_dict(
            {"main1": main1, "main2": main2}, args=["main1"]
        )


def test_wrong_annotation() -> None:
    @dataclasses.dataclass
    class Args:
        x: Union[dict, int] = None  # type: ignore

    with pytest.warns(UserWarning):
        assert tyro.cli(Args, args=[]).x is None
    with pytest.warns(UserWarning):
        assert tyro.cli(Args, args="--x key value k v".split(" ")).x == {
            "key": "value",
            "k": "v",
        }


def test_bad_dict_control() -> None:
    assert tyro.cli(
        Dict[Literal["a", "b"], Literal["c", "d"]], args="a c b d".split(" ")
    ) == {"a": "c", "b": "d"}


def test_bad_dict_key() -> None:
    with pytest.raises(SystemExit):
        tyro.cli(Dict[Literal["a", "b"], Literal["c", "d"]], args="c c d d".split(" "))


def test_bad_dict_val() -> None:
    assert tyro.cli(
        Dict[Literal["a", "b"], Literal["c", "d"]], args="a c b d".split(" ")
    ) == {"a": "c", "b": "d"}
    with pytest.raises(SystemExit):
        tyro.cli(Dict[Literal["a", "b"], Literal["c", "d"]], args="a a b b".split(" "))


def test_invalid_subcommand_default() -> None:
    @dataclasses.dataclass
    class A:
        a: int

    @dataclasses.dataclass
    class B:
        b: int

    @dataclasses.dataclass
    class C:
        c: int

    assert tyro.cli(Union[A, B], args="a --a 5".split(" ")) == A(5)  # type: ignore
    assert tyro.cli(Union[A, B], args="b --b 5".split(" ")) == B(5)  # type: ignore

    with pytest.warns(UserWarning):
        assert tyro.cli(Union[A, B], default=C(3), args=[]) == C(3)  # type: ignore

    with pytest.warns(UserWarning):
        assert tyro.cli(Union[A, B], default=C(3), args="b --b 5".split(" ")) == B(5)  # type: ignore



import dataclasses
import enum
from typing import Any, Dict, Generic, List, Set, Tuple, TypeVar

import pytest

import tyro
from tyro.constructors import UnsupportedTypeAnnotationError


@dataclasses.dataclass(frozen=True)
class Color:
    r: int
    g: int
    b: int


def test_nested_tuple_fixed_single() -> None:
    def main(x: Tuple[Color]) -> Any:
        return x

    assert tyro.cli(main, args="--x.0.r 255 --x.0.g 127 --x.0.b 5".split(" ")) == (
        Color(255, 127, 5),
    )


def test_nested_tuple_fixed_two() -> None:
    def main(x: Tuple[Color, Color]) -> Any:
        return x

    assert tyro.cli(
        main,
        args=(
            "--x.0.r 255 --x.0.g 127 --x.0.b 5 --x.1.r 255 --x.1.g 127 --x.1.b 0"
        ).split(" "),
    ) == (
        Color(255, 127, 5),
        Color(255, 127, 0),
    )


def test_nested_tuple_fixed_three() -> None:
    def main(x: Tuple[Color, int, Color]) -> Any:
        return x

    assert tyro.cli(
        main,
        args=(
            "--x.0.r 255 --x.0.g 127 --x.0.b 5 --x.1 94709 --x.2.r 255 --x.2.g 127"
            " --x.2.b 0"
        ).split(" "),
    ) == (
        Color(255, 127, 5),
        94709,
        Color(255, 127, 0),
    )


def test_nested_tuple_recursive() -> None:
    def main(x: Tuple[Color, Tuple[Color, Color]]) -> Any:
        return x

    assert tyro.cli(
        main,
        args=(
            "--x.0.r 255 --x.0.g 127 --x.0.b 5 --x.1.0.r 255 --x.1.0.g 127 --x.1.0.b 0"
            " --x.1.1.r 255 --x.1.1.g 127 --x.1.1.b 0"
        ).split(" "),
    ) == (
        Color(255, 127, 5),
        (
            Color(255, 127, 0),
            Color(255, 127, 0),
        ),
    )


def test_tuple_bad() -> None:
    # Unable to infer input length.
    def main(x: Tuple[Color, ...]) -> None:
        pass

    with pytest.raises(UnsupportedTypeAnnotationError):
        tyro.cli(main, args=[])


def test_set_bad() -> None:
    # Unable to infer input length.
    def main(x: Set[Color]) -> None:
        pass

    with pytest.raises(UnsupportedTypeAnnotationError):
        tyro.cli(main, args=[])


def test_set_ok() -> None:
    def main(x: Set[Color] = {Color(255, 0, 0)}) -> Any:
        return x

    assert tyro.cli(main, args=[]) == {Color(255, 0, 0)}
    assert tyro.cli(main, args="--x.0.r 127".split(" ")) == {Color(127, 0, 0)}


def test_list_bad() -> None:
    # Unable to infer input length.
    def main(x: List[Color]) -> None:
        pass

    with pytest.raises(UnsupportedTypeAnnotationError):
        tyro.cli(main, args=[])


def test_list_ok() -> None:
    def main(x: List[Color] = [Color(255, 0, 0)]) -> Any:
        return x

    assert tyro.cli(main, args=[]) == [Color(255, 0, 0)]
    assert tyro.cli(main, args="--x.0.r 127".split(" ")) == [Color(127, 0, 0)]


def test_list_object() -> None:
    def main(x: List[object] = [Color(255, 0, 0)]) -> Any:
        return x

    assert tyro.cli(main, args=[]) == [Color(255, 0, 0)]
    assert tyro.cli(main, args="--x.0.r 127".split(" ")) == [Color(127, 0, 0)]


def test_list_any() -> None:
    def main(x: List[Any] = [Color(255, 0, 0)]) -> Any:
        return x

    assert tyro.cli(main, args=[]) == [Color(255, 0, 0)]
    assert tyro.cli(main, args="--x.0.r 127".split(" ")) == [Color(127, 0, 0)]


def test_tuple_in_list() -> None:
    def main(x: List[Tuple[Color]] = [(Color(255, 0, 0),)]) -> Any:
        return x

    assert tyro.cli(main, args=[]) == [(Color(255, 0, 0),)]
    assert tyro.cli(main, args="--x.0.0.r 127".split(" ")) == [(Color(127, 0, 0),)]


def test_tuple_variable() -> None:
    def main(x: Tuple[Color, ...] = (Color(255, 0, 0), Color(255, 0, 127))) -> Any:
        return x

    assert tyro.cli(main, args=[]) == (Color(255, 0, 0), Color(255, 0, 127))
    assert tyro.cli(main, args="--x.0.r 127".split(" ")) == (
        Color(127, 0, 0),
        Color(255, 0, 127),
    )


def test_dict_bad() -> None:
    def main(x: Dict[str, Color]) -> Any:
        return x

    with pytest.raises(UnsupportedTypeAnnotationError):
        tyro.cli(main, args=[])


def test_dict_ok() -> None:
    def main(
        x: Dict[str, Color] = {
            "red": Color(255, 0, 0),
            "green": Color(0, 255, 0),
            "blue": Color(0, 0, 255),
        },
    ) -> Any:
        return x

    assert tyro.cli(main, args=[])["green"] == Color(0, 255, 0)
    assert tyro.cli(main, args="--x.green.g 127".split(" "))["green"] == Color(
        0, 127, 0
    )


def test_dict_key_int() -> None:
    def main(
        x: Dict[int, Color] = {
            0: Color(255, 0, 0),
            1: Color(0, 255, 0),
            2: Color(0, 0, 255),
        },
    ) -> Any:
        return x

    assert tyro.cli(main, args=[])[1] == Color(0, 255, 0)
    assert tyro.cli(main, args="--x.1.g 127".split(" "))[1] == Color(0, 127, 0)


def test_dict_key_enum() -> None:
    class ColorType(enum.Enum):
        RED = enum.auto()
        GREEN = enum.auto()
        BLUE = enum.auto()

    def main(
        x: Dict[ColorType, Color] = {
            ColorType.RED: Color(255, 0, 0),
            ColorType.GREEN: Color(0, 255, 0),
            ColorType.BLUE: Color(0, 0, 255),
        },
    ) -> Any:
        return x

    assert tyro.cli(main, args=[])[ColorType.GREEN] == Color(0, 255, 0)
    assert tyro.cli(main, args="--x.GREEN.g 127".split(" "))[ColorType.GREEN] == Color(
        0, 127, 0
    )


def test_dict_nested() -> None:
    def main(
        x: Dict[str, Tuple[Color, int]] = {
            # For each color: RGB and xterm color code.
            "red": (Color(255, 0, 0), 9),
            "green": (Color(0, 255, 0), 10),
            "blue": (Color(0, 0, 255), 12),
        },
    ) -> Any:
        return x

    assert tyro.cli(main, args=[])["green"] == (Color(0, 255, 0), 10)
    assert tyro.cli(main, args="--x.green.0.g 127 --x.green.1 2".split(" "))[
        "green"
    ] == (Color(0, 127, 0), 2)


def test_generic_in_tuple() -> None:
    ScalarType = TypeVar("ScalarType", int, float)

    @dataclasses.dataclass
    class GenericColor(Generic[ScalarType]):
        r: ScalarType
        g: ScalarType
        b: ScalarType

    def main(x: Tuple[GenericColor[float], GenericColor[int]]) -> Any:
        return x

    assert tyro.cli(
        main,
        args=(
            "--x.0.r 0.5 --x.0.g 0.2 --x.0.b 0.3 --x.1.r 25 --x.1.g 2 --x.1.b 3".split(
                " "
            )
        ),
    ) == (GenericColor(0.5, 0.2, 0.3), GenericColor(25, 2, 3))


def test_generic_in_tuple_with_default() -> None:
    ScalarType = TypeVar("ScalarType", int, float)

    @dataclasses.dataclass
    class GenericColor(Generic[ScalarType]):
        r: ScalarType
        g: ScalarType
        b: ScalarType

    def main(
        x: Tuple[GenericColor[float], GenericColor[int]] = (
            GenericColor(0.5, 0.2, 0.3),
            GenericColor[int](25, 2, 3),  # The subscript should be optional.
        ),
    ) -> Any:
        return x

    assert tyro.cli(
        main,
        args=(
            "--x.0.r 0.5 --x.0.g 0.2 --x.0.b 0.3 --x.1.r 25 --x.1.g 2 --x.1.b 3".split(
                " "
            )
        ),
    ) == (GenericColor(0.5, 0.2, 0.3), GenericColor(25, 2, 3))


def test_generic_in_variable_tuple_with_default() -> None:
    ScalarType = TypeVar("ScalarType", int, float)

    @dataclasses.dataclass
    class GenericColor(Generic[ScalarType]):
        r: ScalarType
        g: ScalarType
        b: ScalarType

    def main(
        x: Tuple[GenericColor, ...] = (
            GenericColor[float](0.5, 0.2, 0.3),
            GenericColor[int](25, 2, 3),
        ),
    ) -> Any:
        return x

    assert tyro.cli(
        main,
        args=(
            "--x.0.r 0.5 --x.0.g 0.9 --x.0.b 0.3 --x.1.r 25 --x.1.g 2 --x.1.b 3".split(
                " "
            )
        ),
    ) == (GenericColor(0.5, 0.9, 0.3), GenericColor(25, 2, 3))


def test_generic_in_dict_with_default() -> None:
    ScalarType = TypeVar("ScalarType", int, float)

    @dataclasses.dataclass
    class GenericColor(Generic[ScalarType]):
        r: ScalarType
        g: ScalarType
        b: ScalarType

    def main(
        x: Dict[str, GenericColor] = {
            "float": GenericColor(0.5, 0.2, 0.3),
            "int": GenericColor[int](25, 2, 3),
        },
    ) -> Any:
        return x

    assert tyro.cli(
        main,
        args="--x.float.g 0.1".split(" "),
    )["float"] == GenericColor(0.5, 0.1, 0.3)
    assert tyro.cli(
        main,
        args="--x.int.g 0".split(" "),
    ) == {"float": GenericColor(0.5, 0.2, 0.3), "int": GenericColor(25, 0, 3)}


def test_generic_in_double_nested_dict_with_default() -> None:
    ScalarType = TypeVar("ScalarType", int, float)

    @dataclasses.dataclass
    class GenericColor(Generic[ScalarType]):
        r: ScalarType
        g: ScalarType
        b: ScalarType

    def main(
        x: Dict[str, Dict[str, GenericColor]] = {
            "hello": {
                "float": GenericColor(0.5, 0.2, 0.3),
                "int": GenericColor[int](25, 2, 3),
            }
        },
    ) -> Any:
        return x

    assert tyro.cli(
        main,
        args="--x.hello.float.g 0.1".split(" "),
    )["hello"]["float"] == GenericColor(0.5, 0.1, 0.3)
    assert tyro.cli(
        main,
        args="--x.hello.int.g 0".split(" "),
    ) == {
        "hello": {"float": GenericColor(0.5, 0.2, 0.3), "int": GenericColor(25, 0, 3)}
    }


def test_double_nested_dict_with_inferred_type() -> None:
    def main(
        x: Dict[str, Any] = {
            "hello": {
                "a": Color(5, 2, 3),
                "b": Color(25, 2, 3),
            }
        },
    ) -> Any:
        return x

    assert tyro.cli(
        main,
        args="--x.hello.a.g 1".split(" "),
    )["hello"]["a"] == Color(5, 1, 3)



"""Decorator-based Subcommands

:func:`tyro.extras.SubcommandApp()` provides a decorator-based API for
subcommands, which is inspired by `click <https://click.palletsprojects.com/>`_.

Usage:

    python 04_decorator_subcommands.py --help
    python 04_decorator_subcommands.py greet --help
    python 04_decorator_subcommands.py greet --name Alice
    python 04_decorator_subcommands.py greet --name Bob --loud
    python 04_decorator_subcommands.py addition --help
    python 04_decorator_subcommands.py addition --a 5 --b 3
"""

from tyro.extras import SubcommandApp

app = SubcommandApp()


@app.command
def greet(name: str, loud: bool = False) -> None:
    """Greet someone."""
    greeting = f"Hello, {name}!"
    if loud:
        greeting = greeting.upper()
    print(greeting)


@app.command(name="addition")
def add(a: int, b: int) -> None:
    """Add two numbers."""
    print(f"{a} + {b} = {a + b}")


if __name__ == "__main__":
    app.cli()



"""Tests for features that are not officially features, but should work.

Includes things like omegaconf.MISSING, attrs, etc, which mostly work but either likely
have corner cases or just seem sketchy.
"""

from typing import Tuple

import omegaconf
import pytest

import tyro
import tyro._strings


def test_omegaconf_missing():
    """Passing in a omegaconf.MISSING default; this will mark an argument as required."""

    def main(
        required_a: int,
        optional: int = 3,
        required_b: int = None,  # type: ignore
    ) -> Tuple[int, int, int]:
        return (required_a, optional, required_b)  # type: ignore

    assert tyro.cli(
        main, args="--required-a 3 --optional 4 --required-b 5".split(" ")
    ) == (3, 4, 5)
    assert tyro.cli(main, args="--required-a 3 --required-b 5".split(" ")) == (
        3,
        3,
        5,
    )

    with pytest.raises(SystemExit):
        tyro.cli(main, args="--required-a 3 --optional 4")
    with pytest.raises(SystemExit):
        tyro.cli(main, args="--required-a 3")

    def main2(
        required_a: int,
        optional: int = 3,
        required_b: int = omegaconf.MISSING,
    ) -> Tuple[int, int, int]:
        return (required_a, optional, required_b)

    assert tyro.cli(
        main2, args="--required-a 3 --optional 4 --required-b 5".split(" ")
    ) == (3, 4, 5)
    assert tyro.cli(main2, args="--required-a 3 --required-b 5".split(" ")) == (
        3,
        3,
        5,
    )

    with pytest.raises(SystemExit):
        tyro.cli(main2, args="--required-a 3 --optional 4")
    with pytest.raises(SystemExit):
        tyro.cli(main2, args="--required-a 3")



# mypy: ignore-errors
#
# We can remove this ignore after: https://peps.python.org/pep-0747/

import dataclasses
from typing import Any, Literal, Type

import pytest

import tyro


def test_union_direct() -> None:
    assert tyro.cli(int | str, args=["5"]) == 5
    assert tyro.cli(int | str, args=["five"]) == "five"


def test_union_basic() -> None:
    def main(x: int | str) -> int | str:
        return x

    assert tyro.cli(main, args=["--x", "5"]) == 5
    assert tyro.cli(main, args=["--x", "6"]) == 6
    assert tyro.cli(main, args=["--x", "five"]) == "five"


def test_union_with_list() -> None:
    def main(x: int | str | list[bool]) -> Any:
        return x

    assert tyro.cli(main, args=["--x", "5"]) == 5
    assert tyro.cli(main, args=["--x", "6"]) == 6
    assert tyro.cli(main, args=["--x", "five"]) == "five"
    assert tyro.cli(main, args=["--x", "True"]) == "True"
    assert tyro.cli(main, args=["--x", "True", "False"]) == [True, False]


def test_union_literal() -> None:
    def main(x: Literal[1, 2] | Literal[3, 4, 5] | str) -> int | str:
        return x

    assert tyro.cli(main, args=["--x", "5"]) == 5
    assert tyro.cli(main, args=["--x", "6"]) == "6"
    assert tyro.cli(main, args=["--x", "five"]) == "five"


def test_super_nested() -> None:
    def main(
        x: (
            None
            | list[
                tuple[
                    None | int,
                    Literal[3, 4],
                    tuple[int, int] | tuple[str, str],
                ]
            ]
        ) = None,
    ) -> Any:
        return x

    assert tyro.cli(main, args=[]) is None
    assert tyro.cli(main, args="--x None".split(" ")) is None
    assert tyro.cli(main, args="--x None 3 2 2".split(" ")) == [(None, 3, (2, 2))]
    assert tyro.cli(main, args="--x 2 3 x 2".split(" ")) == [(2, 3, ("x", "2"))]
    assert tyro.cli(main, args="--x 2 3 x 2 2 3 1 2".split(" ")) == [
        (2, 3, ("x", "2")),
        (2, 3, (1, 2)),
    ]
    with pytest.raises(SystemExit):
        tyro.cli(main, args=["--help"])


def test_type() -> None:
    """Test adapted from mirceamironenco: https://github.com/brentyi/tyro/issues/164"""

    class Thing: ...

    class SubThing(Thing): ...

    @dataclasses.dataclass
    class Config:
        foo: int
        barr: type[Thing] = dataclasses.field(default=SubThing)
        bar: type[Thing] = dataclasses.field(default=SubThing)

    assert tyro.cli(Config, args=["--foo", "5"]) == Config(5, SubThing, SubThing)


def test_type_default_factory() -> None:
    """Test adapted from mirceamironenco: https://github.com/brentyi/tyro/issues/164"""

    @dataclasses.dataclass
    class Config:
        foo: int
        bar: type[Type] = dataclasses.field(default_factory=lambda: Type)

    assert tyro.cli(Config, args=["--foo", "5"]) == Config(5)



from typing import Any


class Singleton:
    # Singleton pattern.
    # https://www.python.org/download/releases/2.2/descrintro/#__new__
    def __new__(cls, *args, **kwds):
        it = cls.__dict__.get("__it__")
        if it is not None:
            return it
        cls.__it__ = it = object.__new__(cls)
        it.init(*args, **kwds)
        return it

    def init(self, *args, **kwds):
        pass


class PropagatingMissingType(Singleton):
    pass


class NonpropagatingMissingType(Singleton):
    pass


class ExcludeFromCallType(Singleton):
    pass


class NotRequiredButWeDontKnowTheValueType(Singleton):
    pass


# We have two types of missing sentinels: a propagating missing value, which when set as
# a default will set all child values of nested structures as missing as well, and a
# nonpropagating missing sentinel, which does not override child defaults.
MISSING: Any = PropagatingMissingType()
"""Sentinel value to mark default values as missing. Can be used to mark fields
passed in via `default=` for `tyro.cli()` as required.

When used, the 'missing' semantics propagate to children. For example, if we write:

.. code-block:: python

    def main(inner: Dataclass = tyro.MISSING) -> None:
        ...

    tyro.cli(main)

then all fields belonging to ``Dataclass`` will be marked as missing, even if a
default exists in the dataclass definition.
"""
MISSING_NONPROP: Any = NonpropagatingMissingType()
"""Non-propagating version of :data:`tyro.MISSING`.

When used, the 'missing' semantics do not propagate to children. For example:

.. code-block:: python

    def main(inner: Dataclass = tyro.constructors.MISSING_NONPROP) -> None:
        ...

    tyro.cli(main)

is equivalent to:

.. code-block:: python

    def main(inner: Dataclass) -> None:
        ...

    tyro.cli(main)

where default values for fields belonging to ``Dataclass`` will be taken from
the dataclass definition.
"""

# When total=False in a TypedDict, we exclude fields from the constructor by default.
NOT_REQUIRED_BUT_WE_DONT_KNOW_THE_VALUE = NotRequiredButWeDontKnowTheValueType()


EXCLUDE_FROM_CALL = ExcludeFromCallType()


MISSING_AND_MISSING_NONPROP = (
    MISSING,
    MISSING_NONPROP,
)
"""Singletons that are considered missing values when generating CLI interfaces."""

DEFAULT_SENTINEL_SINGLETONS = MISSING_AND_MISSING_NONPROP + (
    NOT_REQUIRED_BUT_WE_DONT_KNOW_THE_VALUE,
    EXCLUDE_FROM_CALL,
)



import contextlib
import dataclasses
import io
from typing import Union

import pytest
from typing_extensions import Annotated, Literal, Optional

import tyro


# https://github.com/brentyi/tyro/issues/9
@dataclasses.dataclass(frozen=True)
class Subtype:
    data: int = 1


@dataclasses.dataclass(frozen=True)
class TypeA:
    subtype: Subtype = Subtype(1)


@dataclasses.dataclass(frozen=True)
class TypeB:
    subtype: Subtype = Subtype(2)


@dataclasses.dataclass(frozen=True)
class Wrapper:
    supertype: Union[TypeA, TypeB] = TypeA()


def test_bash():
    target = io.StringIO()
    with pytest.raises(SystemExit), contextlib.redirect_stdout(target):
        tyro.cli(Wrapper, args=["--tyro-print-completion", "bash"])
    assert "# AUTOMATICALLY GENERATED by `shtab`" in target.getvalue()


def test_zsh():
    target = io.StringIO()
    with pytest.raises(SystemExit), contextlib.redirect_stdout(target):
        tyro.cli(Wrapper, args=["--tyro-print-completion", "zsh"])
    assert "# AUTOMATICALLY GENERATED by `shtab`" in target.getvalue()


def test_completion_zsh():
    """https://github.com/brentyi/tyro/issues/158"""

    def start_device(
        preset: Annotated[
            Optional[Literal["rgb", "depth", "ir"]], tyro.conf.arg(aliases=["-p"])
        ] = None,
        frame: Annotated[
            Literal["world", "base"], tyro.conf.arg(aliases=["-f"])
        ] = "world",
    ) -> None:
        """
        Start device with the given preset.

        :param preset: device preset to use.
        :param frame: coordinate frame to use.
        """

    target = io.StringIO()
    with pytest.raises(SystemExit), contextlib.redirect_stdout(target):
        tyro.cli(start_device, args=["--tyro-print-completion", "bash"])

    completion_script = target.getvalue()
    print(completion_script)
    assert "# AUTOMATICALLY GENERATED by `shtab`" in completion_script
    assert "preset_choices=(" in completion_script
    assert "p_choices=(" in completion_script
    assert "frame_choices=(" in completion_script
    assert "f_choices=(" in completion_script



import pytest

from tyro.extras import SubcommandApp

app = SubcommandApp()
app_just_one = SubcommandApp()


@app_just_one.command
@app.command
def greet(name: str, loud: bool = False) -> None:
    """Greet someone."""
    greeting = f"Hello, {name}!"
    if loud:
        greeting = greeting.upper()
    print(greeting)


@app.command(name="addition")
def add(a: int, b: int) -> None:
    """Add two numbers."""
    print(f"{a} + {b} = {a + b}")


def test_app_just_one_cli(capsys):
    # Test: `python my_script.py --help`
    with pytest.raises(SystemExit):
        app_just_one.cli(args=["--help"])
    captured = capsys.readouterr()
    assert "usage: " in captured.out
    assert "greet" not in captured.out
    assert "addition" not in captured.out
    assert "--name" in captured.out


def test_app_cli(capsys):
    # Test: `python my_script.py --help`
    with pytest.raises(SystemExit):
        app.cli(args=["--help"])
    captured = capsys.readouterr()
    assert "usage: " in captured.out
    assert "greet" in captured.out
    assert "addition" in captured.out

    # Test: `python my_script.py greet --help`
    with pytest.raises(SystemExit):
        app.cli(args=["greet", "--help"], sort_subcommands=False)
    captured = capsys.readouterr()
    assert "usage: " in captured.out
    assert "Greet someone." in captured.out

    # Test: `python my_script.py greet --name Alice`
    app.cli(args=["greet", "--name", "Alice"], sort_subcommands=True)
    captured = capsys.readouterr()
    assert captured.out.strip() == "Hello, Alice!"

    # Test: `python my_script.py greet --name Bob --loud`
    app.cli(args=["greet", "--name", "Bob", "--loud"])
    captured = capsys.readouterr()
    assert captured.out.strip() == "HELLO, BOB!"

    # Test: `python my_script.py addition --help`
    with pytest.raises(SystemExit):
        app.cli(args=["addition", "--help"])
    captured = capsys.readouterr()
    assert "usage: " in captured.out
    assert "Add two numbers." in captured.out

    # Test: `python my_script.py addition 5 3`
    app.cli(args=["addition", "--a", "5", "--b", "3"])
    captured = capsys.readouterr()
    assert captured.out.strip() == "5 + 3 = 8"



"""Adapted from @mirceamironenco: https://github.com/brentyi/tyro/issues/170"""

from dataclasses import dataclass

import tyro


class LayerAExample:
    def __init__(self, **kwargs): ...


@dataclass
class LayerAConfig:
    _target: type = LayerAExample
    foo: int = 13


class LayerBExample:
    def __init__(self, **kwargs): ...


@dataclass
class LayerBConfig:
    _target: type = LayerBExample
    bar: int = 13


@dataclass
class BlockConfig:
    layer_a: LayerAConfig
    layer_b: LayerBConfig


def test_repeat_suppressed() -> None:
    assert tyro.cli(
        tyro.conf.OmitArgPrefixes[tyro.conf.SuppressFixed[BlockConfig]],
        args="--foo 14".split(" "),
    ) == BlockConfig(LayerAConfig(foo=14), LayerBConfig())



import dataclasses

import tyro


def test_dataclass_init_var() -> None:
    @dataclasses.dataclass
    class DataclassWithInitVar:
        i: dataclasses.InitVar[int]
        x: str

        def __post_init__(self, i: int) -> None:
            self.x += str(i)

    # We can directly pass a dataclass to `tyro.cli()`:
    assert (
        tyro.cli(
            DataclassWithInitVar,
            args=[
                "--i",
                "5",
                "--x",
                "5",
            ],
        ).x
        == "55"
    )



"""Structures as Function Arguments

Structures can also be used as input to functions.

Usage:

    python ./02_nesting_in_func.py --help
    python ./02_nesting_in_func.py --out-dir /tmp/test1
    python ./02_nesting_in_func.py --out-dir /tmp/test2 --config.seed 4
"""

import dataclasses
import pathlib

import tyro


@dataclasses.dataclass
class OptimizerConfig:
    learning_rate: float = 3e-4
    weight_decay: float = 1e-2


@dataclasses.dataclass
class Config:
    # Optimizer options.
    optimizer: OptimizerConfig

    # Random seed.
    seed: int = 0


def train(
    out_dir: pathlib.Path,
    config: Config,
) -> None:
    """Train a model.

    Args:
        out_dir: Where to save logs and checkpoints.
        config: Experiment configuration.
    """
    print(f"Saving to: {out_dir}")
    print(f"Config: {config}")


if __name__ == "__main__":
    tyro.cli(train)



from __future__ import annotations

import dataclasses
import functools

import pytest
from helptext_utils import get_helptext_with_checks

import tyro


def test_partial_func() -> None:
    def main(a: int, b: str) -> str:
        return b * a

    assert tyro.cli(functools.partial(main, a=3), args=["--b", "hi"]) == "hihihi"


def test_partial_class() -> None:
    class Main:
        def __init__(self, a: int, b: str) -> None:
            self.inner = b * a

    assert tyro.cli(functools.partial(Main, a=3), args=["--b", "hi"]).inner == "hihihi"


def test_partial_helptext_func() -> None:
    def main(a: int, b: str) -> str:
        """Hello!"""
        return b * a

    helptext = get_helptext_with_checks(functools.partial(main, b="hello world"))
    assert "partial" not in helptext
    assert "Hello!" in helptext
    assert "hello world" in helptext


def test_partial_helptext_class() -> None:
    class Main:
        """Hello!"""

        def __init__(self, a: int, b: str) -> None:
            self.inner = b * a

    helptext = get_helptext_with_checks(functools.partial(Main, b="3"))
    assert "partial" not in helptext
    assert "Hello!" in helptext


def test_wraps_func() -> None:
    def main(a: int, b: str) -> str:
        return b * a

    @functools.wraps(main)
    def wrapper(*args, **kwargs) -> int:
        return kwargs["a"]

    assert tyro.cli(wrapper, args=["--a", "3", "--b", "hi"]) == 3
    with pytest.raises(SystemExit):
        tyro.cli(wrapper, args=["--a", "3"])


def test_wraps_partial_func_helptext() -> None:
    def main(a: int, b: str) -> str:
        """Hello!

        Args:
            a: Argument.
            b: Argument.
        """
        return b * a

    @functools.wraps(main)
    def wrapper(*args, **kwargs) -> int:
        return kwargs["a"]

    assert tyro.cli(functools.partial(wrapper, a=3), args=["--b", "hi"]) == 3

    helptext = get_helptext_with_checks(functools.partial(wrapper, b="3"))
    assert "wraps" not in helptext
    assert "Hello!" in helptext
    assert "Argument." in helptext


def test_wraps_partial_class_helptext() -> None:
    class Main:
        """Hello!"""

        def __init__(self, a: int, b: str) -> None:
            self.inner = b * a

    @functools.wraps(Main)
    def wrapper(*args, **kwargs) -> int:
        return kwargs["a"]

    assert tyro.cli(functools.partial(wrapper, a=3), args=["--b", "hi"]) == 3

    helptext = get_helptext_with_checks(functools.partial(wrapper, b="3"))
    assert "wraps" not in helptext
    assert "Hello!" in helptext


@dataclasses.dataclass
class WrappedDataclass:
    """Hello!"""

    a: int
    b: str
    """Second field."""


def test_wraps_partial_dataclass() -> None:
    @functools.wraps(WrappedDataclass)
    def wrapper(*args, **kwargs) -> str:
        return kwargs["a"] * kwargs["b"]

    assert tyro.cli(functools.partial(wrapper, a=3), args=["--b", "hi"]) == "hihihi"
    assert (
        tyro.cli(
            functools.partial(functools.partial(wrapper, a=3), b="hello"),
            args=["--b", "hi"],
        )
        == "hihihi"
    )
    assert (
        tyro.cli(functools.partial(functools.partial(wrapper, a=3), b="hello"), args=[])
        == "hellohellohello"
    )

    helptext = get_helptext_with_checks(functools.partial(wrapper, b="3"))
    assert "wraps" not in helptext
    assert "Hello!" in helptext
    assert "Second field." in helptext



"""Helpers for parsing docstrings. Used for helptext generation."""

import builtins
import collections.abc
import dataclasses
import functools
import inspect
import io
import itertools
import sys
import tokenize
from typing import Callable, Dict, Generic, Hashable, List, Optional, Set, Type, TypeVar

import docstring_parser
from typing_extensions import get_origin, is_typeddict

from . import _resolver, _strings, _unsafe_cache

T = TypeVar("T", bound=Callable)


@dataclasses.dataclass(frozen=True)
class _Token:
    token_type: int
    content: str
    logical_line: int
    actual_line: int


@dataclasses.dataclass(frozen=True)
class _FieldData:
    index: int
    logical_line: int
    actual_line: int
    prev_field_logical_line: int


@dataclasses.dataclass(frozen=True)
class _ClassTokenization:
    tokens: List[_Token]
    tokens_from_logical_line: Dict[int, List[_Token]]
    tokens_from_actual_line: Dict[int, List[_Token]]
    field_data_from_name: Dict[str, _FieldData]

    @staticmethod
    @_unsafe_cache.unsafe_cache(64)
    def make(clz) -> "_ClassTokenization":
        """Parse the source code of a class, and cache some tokenization information."""
        readline = io.BytesIO(inspect.getsource(clz).encode("utf-8")).readline

        tokens: List[_Token] = []
        tokens_from_logical_line: Dict[int, List[_Token]] = {1: []}
        tokens_from_actual_line: Dict[int, List[_Token]] = {1: []}
        field_data_from_name: Dict[str, _FieldData] = {}

        logical_line: int = 1
        actual_line: int = 1
        for toktype, tok, start, end, line in tokenize.tokenize(readline):
            # Note: we only track logical line numbers, which are delimited by
            # `tokenize.NEWLINE`. `tokenize.NL` tokens appear when logical lines are
            # broken into multiple lines of code; these are ignored.
            if toktype == tokenize.NEWLINE:
                logical_line += 1
                actual_line += 1
                tokens_from_logical_line[logical_line] = []
                tokens_from_actual_line[actual_line] = []
            elif toktype == tokenize.NL:
                actual_line += 1
                tokens_from_actual_line[actual_line] = []
            elif toktype is not tokenize.INDENT:
                token = _Token(
                    token_type=toktype,
                    content=tok,
                    logical_line=logical_line,
                    actual_line=actual_line,
                )
                tokens.append(token)
                tokens_from_logical_line[logical_line].append(token)
                tokens_from_actual_line[actual_line].append(token)

        prev_field_logical_line: int = 1
        for i, token in enumerate(tokens[:-1]):
            if token.token_type == tokenize.NAME:
                # Naive heuristic for field names.
                is_first_token = True
                for t in tokens_from_logical_line[token.logical_line]:
                    if t == token:
                        break
                    if t.token_type is not tokenize.COMMENT:
                        is_first_token = False
                        break

                if not is_first_token:
                    continue

                if (
                    tokens[i + 1].content == ":"
                    and token.content not in field_data_from_name
                ):
                    field_data_from_name[token.content] = _FieldData(
                        index=i,
                        logical_line=token.logical_line,
                        actual_line=token.actual_line,
                        prev_field_logical_line=prev_field_logical_line,
                    )
                    prev_field_logical_line = token.logical_line

        return _ClassTokenization(
            tokens=tokens,
            tokens_from_logical_line=tokens_from_logical_line,
            tokens_from_actual_line=tokens_from_actual_line,
            field_data_from_name=field_data_from_name,
        )


@_unsafe_cache.unsafe_cache(1024)
def get_class_tokenization_with_field(
    cls: Type, field_name: str
) -> Optional[_ClassTokenization]:
    # Search for token in this class + all parents.
    found_field: bool = False
    classes_to_search = cls.__mro__
    tokenization = None
    for search_cls in classes_to_search:
        # Inherited generics seem challenging for now.
        # https://github.com/python/typing/issues/777
        assert search_cls is Generic or get_origin(search_cls) is None

        try:
            tokenization = _ClassTokenization.make(search_cls)  # type: ignore
        except OSError:
            # OSError is raised when we can't read the source code. This is
            # fine, we just assume there's no docstring. We can uncomment the
            # assert below for debugging.
            #
            # assert (
            #     # Dynamic dataclasses.
            #     "could not find class definition" in e.args[0]
            #     # Pydantic.
            #     or "source code not available" in e.args[0]
            #     # Third error that can be raised by inspect.py.
            #     or "could not get source code" in e.args[0]
            # )
            return None
        except TypeError as e:  # pragma: no cover
            # Notebooks cause “___ is a built-in class” TypeError.
            assert "built-in class" in e.args[0]
            return None

        # Grab field-specific tokenization data.
        if field_name in tokenization.field_data_from_name:
            found_field = True
            break

    if dataclasses.is_dataclass(cls):
        assert found_field, (
            "Docstring parsing error -- this usually means that there are multiple"
            " dataclasses in the same file with the same name but different scopes."
        )

    return tokenization


@functools.lru_cache(maxsize=1024)
def parse_docstring_from_object(obj: object) -> Dict[str, str]:
    return {
        doc.arg_name: doc.description
        for doc in docstring_parser.parse_from_object(obj).params
        if doc.description is not None
    }


@_unsafe_cache.unsafe_cache(1024)
def get_field_docstring(cls: Type, field_name: str) -> Optional[str]:
    """Get docstring for a field in a class."""

    # NoneType will break docstring_parser.
    if cls is type(None):
        return None

    # Try to parse using docstring_parser.
    for cls_search in cls.__mro__:
        if cls_search.__module__ == "builtins":
            continue  # Skip `object`, `Callable`, `tuple`, etc.
        docstring = parse_docstring_from_object(cls_search).get(field_name, None)
        if docstring is not None:
            return _strings.dedent(
                _strings.remove_single_line_breaks(docstring)
            ).strip()

    # If docstring_parser failed, let's try looking for comments.
    tokenization = get_class_tokenization_with_field(cls, field_name)
    if tokenization is None:  # Currently only happens for dynamic dataclasses.
        return None

    field_data = tokenization.field_data_from_name[field_name]

    # Check for comment on the same line as the field.
    final_token_on_line = tokenization.tokens_from_logical_line[
        field_data.logical_line
    ][-1]
    if final_token_on_line.token_type == tokenize.COMMENT:
        comment: str = final_token_on_line.content
        assert comment.startswith("#")
        if comment.startswith("#:"):  # Sphinx autodoc-style comment.
            # https://www.sphinx-doc.org/en/master/usage/extensions/autodoc.html#directive-autoattribute
            return _strings.remove_single_line_breaks(comment[2:].strip())
        else:
            return _strings.remove_single_line_breaks(comment[1:].strip())

    # Check for comments that come before the field. This is intentionally written to
    # support comments covering multiple (grouped) fields, for example:
    #
    #     # Optimizer hyperparameters.
    #     learning_rate: float
    #     beta1: float
    #     beta2: float
    #
    # In this case, 'Optimizer hyperparameters' will be treated as the docstring for all
    # 3 fields. There are tradeoffs we are making here.
    #
    # The exception this is Sphinx-style comments:
    #
    #     #: The learning rate.
    #     learning_rate: float
    #     beta1: float
    #     beta2: float
    #
    # Where, by convention the comment only applies to the field that directly follows it.

    # Get first line of the class definition, excluding decorators. This logic is only
    # needed for Python >= 3.9; in 3.8, we can simply use
    # `tokenization.tokens[0].logical_line`.
    classdef_logical_line = -1
    for token in tokenization.tokens:
        if token.content == "class":
            classdef_logical_line = token.logical_line
            break
    assert classdef_logical_line != -1

    comments: List[str] = []
    current_actual_line = field_data.actual_line - 1
    directly_above_field = True
    is_sphinx_doc_comment = False
    while current_actual_line in tokenization.tokens_from_actual_line:
        actual_line_tokens = tokenization.tokens_from_actual_line[current_actual_line]

        # We stop looking if we find an empty line.
        if len(actual_line_tokens) == 0:
            break

        # We don't look in the first logical line. This includes all comments that come
        # before the end parentheses in the class definition (eg comments in the
        # subclass list).
        if actual_line_tokens[0].logical_line <= classdef_logical_line:
            break

        # Record single comments!
        if (
            len(actual_line_tokens) == 1
            and actual_line_tokens[0].token_type is tokenize.COMMENT
        ):
            (comment_token,) = actual_line_tokens
            assert comment_token.content.startswith("#")
            if comment_token.content.startswith("#:"):  # Sphinx autodoc-style comment.
                # https://www.sphinx-doc.org/en/master/usage/extensions/autodoc.html#directive-autoattribute
                comments.append(comment_token.content[2:].strip())
                is_sphinx_doc_comment = True
            else:
                comments.append(comment_token.content[1:].strip())
        elif len(comments) > 0:
            # Comments should be contiguous.
            break
        else:
            # This comment is not directly above the current field.
            directly_above_field = False

        current_actual_line -= 1

    if len(comments) > 0 and not (is_sphinx_doc_comment and not directly_above_field):
        return _strings.remove_single_line_breaks("\n".join(reversed(comments)))

    return None


_callable_description_blocklist: Set[Hashable] = set(
    filter(
        lambda x: isinstance(x, Hashable),  # type: ignore
        itertools.chain(
            vars(builtins).values(),
            vars(collections.abc).values(),
        ),
    )
)


@_unsafe_cache.unsafe_cache(1024)
def get_callable_description(f: Callable) -> str:
    """Get description associated with a callable via docstring parsing.

    `dataclasses.dataclass` will automatically populate __doc__ based on the
    fields of the class if a docstring is not specified; this helper will
    ignore these docstrings."""

    f, _ = _resolver.resolve_generic_types(f)
    f = _resolver.unwrap_origin_strip_extras(f)
    if f in _callable_description_blocklist:
        return ""

    # Return original docstring when used with functools.partial, not
    # functools.partial's docstinrg.
    if isinstance(f, functools.partial):
        f = f.func

    if "pydantic" in sys.modules.keys():
        try:
            import pydantic
        except ImportError:
            # Needed for mock import test.
            pydantic = None  # type: ignore
    else:
        pydantic = None  # type: ignore

    # Note inspect.getdoc() causes some corner cases with TypedDicts.
    docstring = f.__doc__
    if (
        docstring is None
        and inspect.isclass(f)
        # Ignore TypedDict's __init__ docstring, because it will just be `dict`
        and not is_typeddict(f)
        # Ignore NamedTuple __init__ docstring.
        and not _resolver.is_namedtuple(f)
        # Ignore pydantic base model constructor docstring.
        and not (pydantic is not None and f.__init__ is pydantic.BaseModel.__init__)  # type: ignore
    ):
        docstring = f.__init__.__doc__  # type: ignore
    if docstring is None:
        return ""

    docstring = _strings.dedent(docstring)

    if dataclasses.is_dataclass(f):
        default_doc = f.__name__ + str(inspect.signature(f)).replace(" -> None", "")  # type: ignore
        if docstring == default_doc:
            return ""

    parsed_docstring = docstring_parser.parse(docstring)

    parts: List[str] = []
    if parsed_docstring.short_description is not None:
        parts.append(parsed_docstring.short_description)
    if parsed_docstring.long_description is not None:
        parts.append(parsed_docstring.long_description)
    return "\n".join(parts)



"""Custom Structs (Registry)

In this example, we use a :class:`tyro.constructors.ConstructorRegistry` to
add support for a custom type.

.. warning::

    This will be complicated!

Usage:
    python ./04_struct_registry.py --help
    python ./04_struct_registry.py --bounds.lower 5 --bounds.upper 10
"""

import tyro


# A custom type that we'll add support for to tyro.
class Bounds:
    def __init__(self, lower: int, upper: int):
        self.bounds = (lower, upper)

    def __repr__(self) -> str:
        return f"(lower={self.bounds[0]}, upper={self.bounds[1]})"


# Create a custom registry, which stores constructor rules.
custom_registry = tyro.constructors.ConstructorRegistry()


# Define a rule that applies to all types that match `Bounds`.
@custom_registry.struct_rule
def _(
    type_info: tyro.constructors.StructTypeInfo,
) -> tyro.constructors.StructConstructorSpec | None:
    # We return `None` if the rule does not apply.
    if type_info.type != Bounds:
        return None

    # We can extract the default value of the field from `type_info`.
    if isinstance(type_info.default, Bounds):
        # If the default value is a `Bounds` instance, we don't need to generate a constructor.
        default = (type_info.default.bounds[0], type_info.default.bounds[1])
        is_default_overridden = True
    else:
        # Otherwise, the default value is missing. We'll mark the child defaults as missing as well.
        assert type_info.default in (
            tyro.constructors.MISSING,
            tyro.constructors.MISSING_NONPROP,
        )
        default = (tyro.MISSING, tyro.MISSING)
        is_default_overridden = False

    # If the rule applies, we return the constructor spec.
    return tyro.constructors.StructConstructorSpec(
        # The instantiate function will be called with the fields as keyword arguments.
        instantiate=Bounds,
        fields=(
            tyro.constructors.StructFieldSpec(
                name="lower",
                type=int,
                default=default[0],
                is_default_overridden=is_default_overridden,
                helptext="Lower bound." "",
            ),
            tyro.constructors.StructFieldSpec(
                name="upper",
                type=int,
                default=default[1],
                is_default_overridden=is_default_overridden,
                helptext="Upper bound." "",
            ),
        ),
    )


def main(
    bounds: Bounds,
    bounds_with_default: Bounds = Bounds(0, 100),
) -> None:
    """A function with two `Bounds` instances as input."""
    print(f"{bounds=}")
    print(f"{bounds_with_default=}")


if __name__ == "__main__":
    # To activate a custom registry, we should use it as a context manager.
    with custom_registry:
        tyro.cli(main)



"""Core functionality for calling functions with arguments specified by argparse
namespaces."""

from __future__ import annotations

import dataclasses
import itertools
from functools import partial
from typing import Any, Callable, Dict, List, Set, Tuple, TypeVar, Union

from typing_extensions import get_args

from . import _arguments, _fields, _parsers, _resolver, _singleton, _strings
from .conf import _markers


@dataclasses.dataclass(frozen=True)
class InstantiationError(Exception):
    """Exception raised when instantiation fail; this typically means that values from
    the CLI are invalid."""

    message: str
    arg: Union[_arguments.ArgumentDefinition, str]


T = TypeVar("T")


def callable_with_args(
    f: Callable[..., T],
    parser_definition: _parsers.ParserSpecification,
    default_instance: Union[T, _singleton.NonpropagatingMissingType],
    value_from_prefixed_field_name: Dict[str, Any],
    field_name_prefix: str,
) -> Tuple[Callable[[], T], Set[str]]:
    """Populate `f` with arguments specified by a dictionary of values from argparse.

    Returns a partialed version of `f` with arguments populated, and a set of
    used arguments.

    We return a `Callable[[], OutT]` instead of `T` directly for aesthetic
    reasons; it lets use reduce layers in stack traces for errors from
    functions passed to `tyro`.
    """

    positional_args: List[Any] = []
    kwargs: Dict[str, Any] = {}
    consumed_keywords: Set[str] = set()

    def get_value_from_arg(prefixed_field_name: str) -> Any:
        """Helper for getting values from `value_from_arg` + doing some extra
        asserts."""
        assert (
            prefixed_field_name in value_from_prefixed_field_name
        ), f"{prefixed_field_name} not in {value_from_prefixed_field_name}"
        return value_from_prefixed_field_name[prefixed_field_name]

    arg_from_prefixed_field_name: Dict[str, _arguments.ArgumentDefinition] = {}
    for arg in parser_definition.args:
        arg_from_prefixed_field_name[
            _strings.make_field_name([arg.intern_prefix, arg.field.intern_name])
        ] = arg

    any_arguments_provided = False

    for field in parser_definition.field_list:
        value: Any
        prefixed_field_name = _strings.make_field_name(
            [field_name_prefix, field.intern_name]
        )

        # Resolve field type.
        field_type = field.type_stripped
        if prefixed_field_name in arg_from_prefixed_field_name:
            assert prefixed_field_name not in consumed_keywords

            # Standard arguments.
            arg = arg_from_prefixed_field_name[prefixed_field_name]
            name_maybe_prefixed = prefixed_field_name
            consumed_keywords.add(name_maybe_prefixed)
            if not arg.lowered.is_fixed():
                value = get_value_from_arg(name_maybe_prefixed)

                if value in _fields.MISSING_AND_MISSING_NONPROP:
                    value = arg.field.default

                    # Consider a function with a positional sequence argument:
                    #
                    #     def f(x: tuple[int, ...], /)
                    #
                    # If we run this script with no arguments, we should interpret this
                    # as empty input for x. But the argparse default will be a MISSING
                    # value, and the field default will be inspect.Parameter.empty.
                    if (
                        value in _fields.MISSING_AND_MISSING_NONPROP
                        and field.is_positional_call()
                        and arg.lowered.nargs in ("?", "*")
                    ):
                        value = []
                else:
                    any_arguments_provided = True
                    if arg.lowered.nargs == "?":
                        # Special case for optional positional arguments: this is the
                        # only time that arguments don't come back as a list.
                        value = [value]

                    try:
                        assert arg.lowered.instance_from_str is not None
                        value = arg.lowered.instance_from_str(value)
                    except (ValueError, TypeError) as e:
                        raise InstantiationError(
                            e.args[0],
                            arg,
                        )
            else:
                assert arg.field.default not in _fields.MISSING_AND_MISSING_NONPROP
                value = arg.field.default
                parsed_value = value_from_prefixed_field_name.get(prefixed_field_name)
                if parsed_value not in _fields.MISSING_AND_MISSING_NONPROP:
                    raise InstantiationError(
                        f"{arg.lowered.name_or_flag} was passed in, but"
                        " is a fixed argument that cannot be parsed",
                        arg,
                    )
        elif prefixed_field_name in parser_definition.child_from_prefix:
            # Nested callable.
            if _resolver.unwrap_origin_strip_extras(field_type) is Union:
                field_type = type(field.default)
            get_value, consumed_keywords_child = callable_with_args(
                field_type,
                parser_definition.child_from_prefix[prefixed_field_name],
                field.default,
                value_from_prefixed_field_name,
                field_name_prefix=prefixed_field_name,
            )
            value = get_value()
            del get_value
            consumed_keywords |= consumed_keywords_child
        else:
            # Unions over dataclasses (subparsers). This is the only other option.
            subparser_def = parser_definition.subparsers_from_intern_prefix[
                prefixed_field_name
            ]
            subparser_dest = _strings.make_subparser_dest(name=prefixed_field_name)
            consumed_keywords.add(subparser_dest)
            if subparser_dest in value_from_prefixed_field_name:
                subparser_name = get_value_from_arg(subparser_dest)
            else:
                assert (
                    subparser_def.default_instance
                    not in _fields.MISSING_AND_MISSING_NONPROP
                )
                subparser_name = None

            if subparser_name is None:
                # No subparser selected -- this should only happen when we have a
                # default/default_factory set.
                assert (
                    type(None) in get_args(field_type)
                    or subparser_def.default_instance is not None
                )
                value = subparser_def.default_instance
            else:
                chosen_f = subparser_def.options[
                    list(subparser_def.parser_from_name.keys()).index(subparser_name)
                ]
                get_value, consumed_keywords_child = callable_with_args(
                    chosen_f,
                    subparser_def.parser_from_name[subparser_name],
                    (
                        field.default
                        if type(field.default) is chosen_f
                        else _singleton.MISSING_NONPROP
                    ),
                    value_from_prefixed_field_name,
                    field_name_prefix=prefixed_field_name,
                )
                value = get_value()
                del get_value
                consumed_keywords |= consumed_keywords_child

        if value is _singleton.EXCLUDE_FROM_CALL:
            continue

        if _markers._UnpackArgsCall in field.markers:
            if len(positional_args) == 0 and len(kwargs) > 0:
                positional_args.extend(kwargs.values())
                kwargs.clear()
            assert isinstance(value, tuple)
            positional_args.extend(value)
        elif _markers._UnpackKwargsCall in field.markers:
            assert isinstance(value, dict)
            kwargs.update(value)
        elif field.is_positional_call():
            assert len(kwargs) == 0
            positional_args.append(value)
        else:
            kwargs[field.call_argname] = value

    # Logic for _markers._OPTIONAL_GROUP.
    is_missing_list = [
        any(v is m for m in _fields.MISSING_AND_MISSING_NONPROP)
        for v in itertools.chain(positional_args, kwargs.values())
    ]
    if any(is_missing_list):
        if not any_arguments_provided:
            # No arguments were provided in this group.
            return lambda: default_instance, consumed_keywords  # type: ignore

        message = "either all arguments must be provided or none of them."
        if len(kwargs) > 0:
            missing_args: List[str] = []
            for k, v in kwargs.items():
                if v not in _fields.MISSING_AND_MISSING_NONPROP:
                    break

                # Argument is missing.
                found = False
                for arg in arg_from_prefixed_field_name.values():
                    if arg.field.call_argname == k:
                        missing_args.append(arg.lowered.name_or_flag)
                        found = True
                        break
                assert found, "This is likely a bug in tyro."

            if len(missing_args):
                message += f" We're missing arguments {missing_args}."
        raise InstantiationError(
            message,
            field_name_prefix,
        )

    # Note: we unwrap types both before and after narrowing. This is because narrowing
    # sometimes produces types like `Tuple[T1, T2, ...]`, where we actually want just
    # `tuple`.
    unwrapped_f = f
    unwrapped_f = _resolver.swap_type_using_confstruct(unwrapped_f)
    unwrapped_f = _resolver.unwrap_origin_strip_extras(unwrapped_f)
    unwrapped_f = _resolver.narrow_subtypes(unwrapped_f, default_instance)
    unwrapped_f = _resolver.unwrap_origin_strip_extras(unwrapped_f)

    if unwrapped_f in (tuple, list, set):
        if len(positional_args) > 0:
            # Triggered when support_single_arg_types=True is used.
            assert len(kwargs) == 0
            assert len(positional_args) == 1
            return lambda: positional_args[0], consumed_keywords  # type: ignore
        else:
            assert len(positional_args) == 0
            return partial(unwrapped_f, kwargs.values()), consumed_keywords  # type: ignore
    elif unwrapped_f is dict:
        if len(positional_args) > 0:
            # Triggered when support_single_arg_types=True is used.
            assert len(kwargs) == 0
            assert len(positional_args) == 1
            return partial(unwrapped_f, *positional_args), consumed_keywords  # type: ignore
        else:
            assert len(positional_args) == 0
            return lambda: kwargs, consumed_keywords  # type: ignore
    else:
        if field_name_prefix == "":
            # Don't catch any errors for the "root" field. If main() in tyro.cli(main)
            # raises a ValueError, this shouldn't be caught.
            return partial(unwrapped_f, *positional_args, **kwargs), consumed_keywords  # type: ignore
        else:
            # Try to catch ValueErrors raised by field constructors.
            def with_instantiation_error():
                try:
                    return unwrapped_f(*positional_args, **kwargs)
                # If unwrapped_f raises a ValueError, wrap the message with a more informative
                # InstantiationError if possible.
                except ValueError as e:
                    raise InstantiationError(
                        e.args[0],
                        field_name_prefix,
                    )

            return with_instantiation_error, consumed_keywords  # type: ignore



from typing import Any, Callable, Tuple

import torch
from helptext_utils import get_helptext_with_checks
from torch import nn

import tyro


def test_torch_device() -> None:
    def main(device: torch.device) -> torch.device:
        return device

    assert tyro.cli(main, args=["--device", "cpu"]) == torch.device("cpu")


def test_supports_inference_mode_decorator() -> None:
    @torch.inference_mode()
    def main(x: int, device: str) -> Tuple[int, str]:
        return x, device

    assert tyro.cli(main, args="--x 3 --device cuda".split(" ")) == (3, "cuda")


def test_torch_device_2() -> None:
    assert tyro.cli(torch.device, args=["cpu"]) == torch.device("cpu")


def test_unparsable() -> None:
    class Struct:
        a: int = 5
        b: str = "7"

    def main(x: Any = Struct()):
        pass

    helptext = get_helptext_with_checks(main)
    assert "--x {fixed}" not in helptext

    def main2(x: Callable = nn.ReLU):
        pass

    helptext = get_helptext_with_checks(main2)
    assert "--x {fixed}" in helptext
    assert "(fixed to:" in helptext
    assert "torch" in helptext
